from typing import Any, Sequence

import PIL.Image

import pixeltable.exceptions as excs
import pixeltable.type_system as ts
from pixeltable.iterators.base import ComponentIterator


class TileIterator(ComponentIterator):
    """
    Iterator over tiles of an image. Each image will be divided into tiles of size `tile_size`, and the tiles will be
    iterated over in row-major order (left-to-right, then top-to-bottom). An optional `overlap` parameter may be
    specified. If the tiles do not exactly cover the image, then the rightmost and bottommost tiles will be padded with
    blackspace, so that the output images all have the exact size `tile_size`.

    Args:
        image: Image to split into tiles.
        tile_size: Size of each tile, as a pair of integers `[width, height]`.
        overlap: Amount of overlap between adjacent tiles, as a pair of integers `[width, height]`.
    """

    __image: PIL.Image.Image
    __tile_size: Sequence[int]
    __overlap: Sequence[int]
    __width: int
    __height: int
    __xlen: int
    __ylen: int
    __i: int
    __j: int

    def __init__(self, image: PIL.Image.Image, *, tile_size: tuple[int, int], overlap: tuple[int, int] = (0, 0)):
        assert overlap[0] < tile_size[0] and overlap[1] < tile_size[1]

        self.__image = image
        self.__image.load()
        self.__tile_size = tile_size
        self.__overlap = overlap
        self.__width, self.__height = image.size
        # Justification for this formula: let t = tile_size[0], o = overlap[0]. Then the values of w (= width) that
        # exactly accommodate an integer number of tiles are t, 2t - o, 3t - 2o, 4t - 3o, ...
        # This formula ensures that t, 2t - o, 3t - 2o, ... result in an xlen of 1, 2, 3, ...
        # but t + 1, 2t - o + 1, 3t - 2o + 1, ... result in an xlen of 2, 3, 4, ...
        self.__xlen = (self.__width - overlap[0] - 1) // (tile_size[0] - overlap[0]) + 1
        self.__ylen = (self.__height - overlap[1] - 1) // (tile_size[1] - overlap[1]) + 1
        self.__i = 0
        self.__j = 0

    def __next__(self) -> dict[str, Any]:
        if self.__j >= self.__ylen:
            raise StopIteration

        x1 = self.__i * (self.__tile_size[0] - self.__overlap[0])
        y1 = self.__j * (self.__tile_size[1] - self.__overlap[1])
        # If x2 > self.__width, PIL does the right thing and pads the image with blackspace
        x2 = x1 + self.__tile_size[0]
        y2 = y1 + self.__tile_size[1]
        tile = self.__image.crop((x1, y1, x2, y2))
        result = {'tile': tile, 'tile_coord': [self.__i, self.__j], 'tile_box': [x1, y1, x2, y2]}

        self.__i += 1
        if self.__i >= self.__xlen:
            self.__i = 0
            self.__j += 1
        return result

    def close(self) -> None:
        pass

    def set_pos(self, pos: int) -> None:
        self.__j = pos // self.__xlen
        self.__i = pos % self.__xlen

    @classmethod
    def input_schema(cls, *args: Any, **kwargs: Any) -> dict[str, ts.ColumnType]:
        return {'image': ts.ImageType(), 'tile_size': ts.JsonType(), 'overlap': ts.JsonType()}

    @classmethod
    def output_schema(cls, *args: Any, **kwargs: Any) -> tuple[dict[str, ts.ColumnType], list[str]]:
        tile_size = kwargs.get('tile_size')
        overlap = kwargs.get('overlap', (0, 0))
        if overlap[0] >= tile_size[0] or overlap[1] >= tile_size[1]:
            raise excs.Error(f'overlap dimensions {overlap} are not strictly smaller than tile size {tile_size}')
        return {'tile': ts.ImageType(), 'tile_coord': ts.JsonType(), 'tile_box': ts.JsonType()}, ['tile']



from __future__ import annotations

import enum
import logging
from typing import TYPE_CHECKING, Any, Literal, Optional, overload
from uuid import UUID

import pixeltable as pxt
from pixeltable import exceptions as excs, type_system as ts
from pixeltable.env import Env
from pixeltable.utils.filecache import FileCache

from .globals import MediaValidation
from .table import Table
from .table_version import TableVersion
from .table_version_handle import TableVersionHandle
from .table_version_path import TableVersionPath
from .update_status import UpdateStatus

if TYPE_CHECKING:
    from pixeltable import exprs
    from pixeltable.globals import TableDataSource
    from pixeltable.io.table_data_conduit import TableDataConduit

_logger = logging.getLogger('pixeltable')


class OnErrorParameter(enum.Enum):
    """Supported values for the on_error parameter"""

    ABORT = 'abort'
    IGNORE = 'ignore'

    @classmethod
    def is_valid(cls, v: Any) -> bool:
        if isinstance(v, str):
            return v.lower() in [c.value for c in cls]
        return False

    @classmethod
    def fail_on_exception(cls, v: Any) -> bool:
        if not cls.is_valid(v):
            raise ValueError(f'Invalid value for on_error: {v}')
        if isinstance(v, str):
            return v.lower() != cls.IGNORE.value
        return True


class InsertableTable(Table):
    """A `Table` that allows inserting and deleting rows."""

    def __init__(self, dir_id: UUID, tbl_version: TableVersionHandle):
        tbl_version_path = TableVersionPath(tbl_version)
        super().__init__(tbl_version.id, dir_id, tbl_version.get().name, tbl_version_path)
        self._tbl_version = tbl_version

    def _display_name(self) -> str:
        assert not self._tbl_version_path.is_replica()
        return 'table'

    @classmethod
    def _create(
        cls,
        dir_id: UUID,
        name: str,
        schema: dict[str, ts.ColumnType],
        df: Optional[pxt.DataFrame],
        primary_key: list[str],
        num_retained_versions: int,
        comment: str,
        media_validation: MediaValidation,
    ) -> InsertableTable:
        columns = cls._create_columns(schema)
        cls._verify_schema(columns)
        column_names = [col.name for col in columns]
        for pk_col in primary_key:
            if pk_col not in column_names:
                raise excs.Error(f'Primary key column {pk_col!r} not found in table schema.')
            col = columns[column_names.index(pk_col)]
            if col.col_type.nullable:
                raise excs.Error(f'Primary key column {pk_col!r} cannot be nullable.')
            col.is_pk = True

        _, tbl_version = TableVersion.create(
            dir_id,
            name,
            columns,
            num_retained_versions=num_retained_versions,
            comment=comment,
            media_validation=media_validation,
        )
        tbl = cls(dir_id, TableVersionHandle.create(tbl_version))
        # TODO We need to commit before doing the insertion, in order to avoid a primary key (version) collision
        #   when the table metadata gets updated. Once we have a notion of user-defined transactions in
        #   Pixeltable, we can wrap the create/insert in a transaction to avoid this.
        session = Env.get().session
        session.commit()
        if df is not None:
            # A DataFrame was provided, so insert its contents into the table
            # (using the same DB session as the table creation)
            tbl_version.insert(None, df, fail_on_exception=True)
        session.commit()

        _logger.info(f'Created table {name!r}, id={tbl_version.id}')
        Env.get().console_logger.info(f'Created table {name!r}.')
        return tbl

    def _get_metadata(self) -> dict[str, Any]:
        md = super()._get_metadata()
        md['base'] = None
        md['is_view'] = False
        md['is_snapshot'] = False
        return md

    @overload
    def insert(
        self,
        source: Optional[TableDataSource] = None,
        /,
        *,
        source_format: Optional[Literal['csv', 'excel', 'parquet', 'json']] = None,
        schema_overrides: Optional[dict[str, ts.ColumnType]] = None,
        on_error: Literal['abort', 'ignore'] = 'abort',
        print_stats: bool = False,
        **kwargs: Any,
    ) -> UpdateStatus: ...

    @overload
    def insert(
        self, /, *, on_error: Literal['abort', 'ignore'] = 'abort', print_stats: bool = False, **kwargs: Any
    ) -> UpdateStatus: ...

    def insert(
        self,
        source: Optional[TableDataSource] = None,
        /,
        *,
        source_format: Optional[Literal['csv', 'excel', 'parquet', 'json']] = None,
        schema_overrides: Optional[dict[str, ts.ColumnType]] = None,
        on_error: Literal['abort', 'ignore'] = 'abort',
        print_stats: bool = False,
        **kwargs: Any,
    ) -> UpdateStatus:
        from pixeltable.catalog import Catalog
        from pixeltable.io.table_data_conduit import UnkTableDataConduit

        with Catalog.get().begin_xact(tbl=self._tbl_version_path, for_write=True, lock_mutable_tree=True):
            table = self
            if source is None:
                source = [kwargs]
                kwargs = None

            tds = UnkTableDataConduit(
                source, source_format=source_format, src_schema_overrides=schema_overrides, extra_fields=kwargs
            )
            data_source = tds.specialize()
            if data_source.source_column_map is None:
                data_source.src_pk = []

            assert isinstance(table, Table)
            data_source.add_table_info(table)
            data_source.prepare_for_insert_into_table()

            fail_on_exception = OnErrorParameter.fail_on_exception(on_error)
            return table.insert_table_data_source(
                data_source=data_source, fail_on_exception=fail_on_exception, print_stats=print_stats
            )

    def insert_table_data_source(
        self, data_source: TableDataConduit, fail_on_exception: bool, print_stats: bool = False
    ) -> pxt.UpdateStatus:
        """Insert row batches into this table from a `TableDataConduit`."""
        from pixeltable.catalog import Catalog
        from pixeltable.io.table_data_conduit import DFTableDataConduit

        with Catalog.get().begin_xact(tbl=self._tbl_version_path, for_write=True, lock_mutable_tree=True):
            if isinstance(data_source, DFTableDataConduit):
                status = pxt.UpdateStatus()
                status += self._tbl_version.get().insert(
                    rows=None, df=data_source.pxt_df, print_stats=print_stats, fail_on_exception=fail_on_exception
                )
            else:
                status = pxt.UpdateStatus()
                for row_batch in data_source.valid_row_batch():
                    status += self._tbl_version.get().insert(
                        rows=row_batch, df=None, print_stats=print_stats, fail_on_exception=fail_on_exception
                    )

        Env.get().console_logger.info(status.insert_msg)

        FileCache.get().emit_eviction_warnings()
        return status

    def _validate_input_rows(self, rows: list[dict[str, Any]]) -> None:
        """Verify that the input rows match the table schema"""
        valid_col_names = set(self._get_schema().keys())
        reqd_col_names = set(self._tbl_version_path.tbl_version.get().get_required_col_names())
        computed_col_names = set(self._tbl_version_path.tbl_version.get().get_computed_col_names())
        for row in rows:
            assert isinstance(row, dict)
            col_names = set(row.keys())
            if len(reqd_col_names - col_names) > 0:
                raise excs.Error(f'Missing required column(s) ({", ".join(reqd_col_names - col_names)}) in row {row}')

            for col_name, val in row.items():
                if col_name not in valid_col_names:
                    raise excs.Error(f'Unknown column name {col_name!r} in row {row}')
                if col_name in computed_col_names:
                    raise excs.Error(f'Value for computed column {col_name!r} in row {row}')

                # validate data
                col = self._tbl_version_path.get_column(col_name)
                try:
                    # basic sanity checks here
                    checked_val = col.col_type.create_literal(val)
                    row[col_name] = checked_val
                except TypeError as e:
                    msg = str(e)
                    raise excs.Error(f'Error in column {col.name}: {msg[0].lower() + msg[1:]}\nRow: {row}') from e

    def delete(self, where: Optional['exprs.Expr'] = None) -> UpdateStatus:
        """Delete rows in this table.

        Args:
            where: a predicate to filter rows to delete.

        Examples:
            Delete all rows in a table:

            >>> tbl.delete()

            Delete all rows in a table where column `a` is greater than 5:

            >>> tbl.delete(tbl.a > 5)
        """
        from pixeltable.catalog import Catalog

        with Catalog.get().begin_xact(tbl=self._tbl_version_path, for_write=True, lock_mutable_tree=True):
            return self._tbl_version.get().delete(where=where)

    def _get_base_table(self) -> Optional['Table']:
        return None

    @property
    def _effective_base_versions(self) -> list[Optional[int]]:
        return []

    def _table_descriptor(self) -> str:
        return self._display_str()



from __future__ import annotations

from typing import Optional

import sqlalchemy as sql

import pixeltable.type_system as ts

from .data_row import DataRow
from .expr import Expr
from .row_builder import RowBuilder
from .sql_element_cache import SqlElementCache


class IsNull(Expr):
    def __init__(self, e: Expr):
        super().__init__(ts.BoolType())
        self.components = [e]
        self.id = self._create_id()

    def __repr__(self) -> str:
        return f'{self.components[0]} == None'

    def _equals(self, other: IsNull) -> bool:
        return True

    def sql_expr(self, sql_elements: SqlElementCache) -> Optional[sql.ColumnElement]:
        e = sql_elements.get(self.components[0])
        if e is None:
            return None
        return e == None

    def eval(self, data_row: DataRow, row_builder: RowBuilder) -> None:
        data_row[self.slot_idx] = data_row[self.components[0].slot_idx] is None

    @classmethod
    def _from_dict(cls, d: dict, components: list[Expr]) -> IsNull:
        assert len(components) == 1
        return cls(components[0])



"""
Pixeltable [UDFs](https://pixeltable.readme.io/docs/user-defined-functions-udfs) for mathematical operations.

Example:
```python
import pixeltable as pxt

t = pxt.get_table(...)
t.select(t.float_col.floor()).collect()
```
"""

import builtins
import math
from typing import Optional

import sqlalchemy as sql

import pixeltable as pxt
from pixeltable.utils.code import local_public_names


@pxt.udf(is_method=True)
def abs(self: float) -> float:
    """
    Return the absolute value of the given number.

    Equivalent to Python [`builtins.abs()`](https://docs.python.org/3/library/functions.html#abs).
    """
    return builtins.abs(self)


@abs.to_sql
def _(self: sql.ColumnElement) -> sql.ColumnElement:
    return sql.func.abs(self)


@pxt.udf(is_method=True)
def ceil(self: float) -> float:
    """
    Return the ceiling of the given number.

    Equivalent to Python [`float(math.ceil(self))`](https://docs.python.org/3/library/math.html#math.ceil) if `self`
    is finite, or `self` itself if `self` is infinite. (This is slightly different from the default behavior of
    `math.ceil(self)`, which always returns an `int` and raises an error if `self` is infinite. The behavior in
    Pixeltable generalizes the Python operator and is chosen to align with the SQL standard.)
    """
    # This ensures the same behavior as SQL
    if math.isfinite(self):
        return float(math.ceil(self))
    else:
        return self


@ceil.to_sql
def _(self: sql.ColumnElement) -> sql.ColumnElement:
    return sql.func.ceiling(self)


@pxt.udf(is_method=True)
def floor(self: float) -> float:
    """
    Return the ceiling of the given number.

    Equivalent to Python [`float(math.floor(self))`](https://docs.python.org/3/library/math.html#math.ceil) if `self`
    is finite, or `self` itself if `self` is infinite. (This is slightly different from the default behavior of
    `math.floor(self)`, which always returns an `int` and raises an error if `self` is infinite. The behavior of
    Pixeltable generalizes the Python operator and is chosen to align with the SQL standard.)
    """
    # This ensures the same behavior as SQL
    if math.isfinite(self):
        return float(math.floor(self))
    else:
        return self


@floor.to_sql
def _(self: sql.ColumnElement) -> sql.ColumnElement:
    return sql.func.floor(self)


@pxt.udf(is_method=True)
def round(self: float, digits: Optional[int] = None) -> float:
    """
    Round a number to a given precision in decimal digits.

    Equivalent to Python [`builtins.round(self, digits or 0)`](https://docs.python.org/3/library/functions.html#round).
    Note that if `digits` is not specified, the behavior matches `builtins.round(self, 0)` rather than
    `builtins.round(self)`; this ensures that the return type is always `float` (as in SQL) rather than `int`.
    """
    # Set digits explicitly to 0 to guarantee a return type of float; this ensures the same behavior as SQL
    return builtins.round(self, digits or 0)


@round.to_sql
def _(self: sql.ColumnElement, digits: Optional[sql.ColumnElement] = None) -> sql.ColumnElement:
    if digits is None:
        return sql.func.round(self)
    else:
        return sql.func.round(sql.cast(self, sql.Numeric), sql.cast(digits, sql.Integer))


@pxt.udf(is_method=True)
def pow(self: int, other: int) -> float:
    """
    Raise `self` to the power of `other`.

    Equivalent to Python [`self ** other`](https://docs.python.org/3/library/functions.html#pow).
    """
    return self**other


@pow.to_sql
def _(self: sql.ColumnElement, other: sql.ColumnElement) -> sql.ColumnElement:
    return sql.func.pow(self, other)


@pxt.udf(is_method=True)
def bitwise_and(self: int, other: int) -> int:
    """
    Bitwise AND of two integers.

    Equivalent to Python
    [`self & other`](https://docs.python.org/3/library/stdtypes.html#bitwise-operations-on-integer-types).
    """
    return self & other


@bitwise_and.to_sql
def _(self: sql.ColumnElement, other: sql.ColumnElement) -> sql.ColumnElement:
    return self.bitwise_and(other)


@pxt.udf(is_method=True)
def bitwise_or(self: int, other: int) -> int:
    """
    Bitwise OR of two integers.

    Equivalent to Python
    [`self | other`](https://docs.python.org/3/library/stdtypes.html#bitwise-operations-on-integer-types).
    """
    return self | other


@bitwise_or.to_sql
def _(self: sql.ColumnElement, other: sql.ColumnElement) -> sql.ColumnElement:
    return self.bitwise_or(other)


@pxt.udf(is_method=True)
def bitwise_xor(self: int, other: int) -> int:
    """
    Bitwise XOR of two integers.

    Equivalent to Python
    [`self ^ other`](https://docs.python.org/3/library/stdtypes.html#bitwise-operations-on-integer-types).
    """
    return self ^ other


@bitwise_xor.to_sql
def _(self: sql.ColumnElement, other: sql.ColumnElement) -> sql.ColumnElement:
    return self.bitwise_xor(other)


__all__ = local_public_names(__name__)


def __dir__() -> list[str]:
    return __all__



import sys
import urllib.parse
import urllib.request
from pathlib import Path

import requests
from tqdm import tqdm

import pixeltable as pxt
from pixeltable import exceptions as excs
from pixeltable.env import Env
from pixeltable.utils import sha256sum

from .packager import TablePackager, TableRestorer

# These URLs are abstracted out for now, but will be replaced with actual (hard-coded) URLs once the
# pixeltable.com URLs are available.

PIXELTABLE_API_URL = 'https://internal-api.pixeltable.com'


def push_replica(dest_tbl_uri: str, src_tbl: pxt.Table) -> str:
    if not src_tbl._tbl_version.get().is_snapshot:
        raise excs.Error('Only snapshots may be published.')

    packager = TablePackager(src_tbl, additional_md={'table_uri': dest_tbl_uri})
    request_json = packager.md | {'operation_type': 'publish_snapshot'}
    headers_json = {'X-api-key': Env.get().pxt_api_key, 'Content-Type': 'application/json'}
    response = requests.post(PIXELTABLE_API_URL, json=request_json, headers=headers_json)
    if response.status_code != 200:
        raise excs.Error(f'Error publishing snapshot: {response.text}')
    response_json = response.json()
    if not isinstance(response_json, dict) or response_json.get('destination') != 's3':
        raise excs.Error(f'Error publishing snapshot: unexpected response from server.\n{response_json}')
    upload_id = response_json['upload_id']
    destination_uri = response_json['destination_uri']

    Env.get().console_logger.info(f"Creating a snapshot of '{src_tbl._path()}' at: {dest_tbl_uri}")

    bundle = packager.package()

    parsed_location = urllib.parse.urlparse(destination_uri)
    if parsed_location.scheme == 's3':
        _upload_bundle_to_s3(bundle, parsed_location)
    else:
        raise excs.Error(f'Unsupported destination: {destination_uri}')

    Env.get().console_logger.info('Finalizing snapshot ...')

    finalize_request_json = {
        'operation_type': 'finalize_snapshot',
        'upload_id': upload_id,
        'datafile': bundle.name,
        'size': bundle.stat().st_size,
        'sha256': sha256sum(bundle),  # Generate our own SHA for independent verification
    }
    # TODO: Use Pydantic for validation
    finalize_response = requests.post(PIXELTABLE_API_URL, json=finalize_request_json, headers=headers_json)
    if finalize_response.status_code != 200:
        raise excs.Error(f'Error finalizing snapshot: {finalize_response.text}')
    finalize_response_json = finalize_response.json()
    if not isinstance(finalize_response_json, dict) or 'confirmed_table_uri' not in finalize_response_json:
        raise excs.Error(f'Error finalizing snapshot: unexpected response from server.\n{finalize_response_json}')

    confirmed_tbl_uri = finalize_response_json['confirmed_table_uri']
    Env.get().console_logger.info(f'The published snapshot is now available at: {confirmed_tbl_uri}')
    return confirmed_tbl_uri


def _upload_bundle_to_s3(bundle: Path, parsed_location: urllib.parse.ParseResult) -> None:
    from pixeltable.utils.s3 import get_client

    bucket = parsed_location.netloc
    remote_dir = Path(urllib.parse.unquote(urllib.request.url2pathname(parsed_location.path)))
    remote_path = str(remote_dir / bundle.name)[1:]  # Remove initial /

    Env.get().console_logger.info(f'Uploading snapshot to: {bucket}:{remote_path}')

    boto_config = {'max_pool_connections': 5, 'connect_timeout': 15, 'retries': {'max_attempts': 3, 'mode': 'adaptive'}}
    s3_client = get_client(**boto_config)

    upload_args = {'ChecksumAlgorithm': 'SHA256'}

    progress_bar = tqdm(
        desc='Uploading',
        total=bundle.stat().st_size,
        unit='B',
        unit_scale=True,
        unit_divisor=1024,
        miniters=1,  # Update every iteration (should be fine for an upload)
        ncols=100,
        file=sys.stdout,
    )
    s3_client.upload_file(
        Filename=str(bundle), Bucket=bucket, Key=remote_path, ExtraArgs=upload_args, Callback=progress_bar.update
    )


def pull_replica(dest_path: str, src_tbl_uri: str) -> pxt.Table:
    headers_json = {'X-api-key': Env.get().pxt_api_key, 'Content-Type': 'application/json'}
    clone_request_json = {'operation_type': 'clone_snapshot', 'table_uri': src_tbl_uri}
    response = requests.post(PIXELTABLE_API_URL, json=clone_request_json, headers=headers_json)
    if response.status_code != 200:
        raise excs.Error(f'Error cloning snapshot: {response.text}')
    response_json = response.json()
    if not isinstance(response_json, dict) or 'table_uri' not in response_json:
        raise excs.Error(f'Error cloning shapshot: unexpected response from server.\n{response_json}')

    primary_tbl_additional_md = response_json['md']['tables'][0]['table_md']['additional_md']
    bundle_uri = primary_tbl_additional_md['destination_uri']
    bundle_filename = primary_tbl_additional_md['datafile']
    parsed_location = urllib.parse.urlparse(bundle_uri)
    if parsed_location.scheme == 's3':
        bundle_path = _download_bundle_from_s3(parsed_location, bundle_filename)
    else:
        raise excs.Error(f'Unexpected response from server: unsupported bundle uri: {bundle_uri}')

    restorer = TableRestorer(dest_path, response_json)
    tbl = restorer.restore(bundle_path)
    Env.get().console_logger.info(f'Created local replica {tbl._path()!r} from URI: {src_tbl_uri}')
    return tbl


def _download_bundle_from_s3(parsed_location: urllib.parse.ParseResult, bundle_filename: str) -> Path:
    from pixeltable.utils.s3 import get_client

    bucket = parsed_location.netloc
    remote_dir = Path(urllib.parse.unquote(urllib.request.url2pathname(parsed_location.path)))
    remote_path = str(remote_dir / bundle_filename)[1:]  # Remove initial /

    Env.get().console_logger.info(f'Downloading snapshot from: {bucket}:{remote_path}')

    boto_config = {'max_pool_connections': 5, 'connect_timeout': 15, 'retries': {'max_attempts': 3, 'mode': 'adaptive'}}
    s3_client = get_client(**boto_config)

    obj = s3_client.head_object(Bucket=bucket, Key=remote_path)  # Check if the object exists
    bundle_size = obj['ContentLength']

    bundle_path = Path(Env.get().create_tmp_path())
    progress_bar = tqdm(
        desc='Downloading',
        total=bundle_size,
        unit='B',
        unit_scale=True,
        unit_divisor=1024,
        miniters=1,
        ncols=100,
        file=sys.stdout,
    )
    s3_client.download_file(Bucket=bucket, Key=remote_path, Filename=str(bundle_path), Callback=progress_bar.update)
    return bundle_path



"""
Pixeltable [UDFs](https://pixeltable.readme.io/docs/user-defined-functions-udfs)
that wrap various endpoints from the Google Gemini API. In order to use them, you must
first `pip install google-genai` and configure your Gemini credentials, as described in
the [Working with Gemini](https://pixeltable.readme.io/docs/working-with-gemini) tutorial.
"""

import asyncio
import io
from pathlib import Path
from typing import TYPE_CHECKING, Optional

import PIL.Image

import pixeltable as pxt
from pixeltable import env, exceptions as excs, exprs

if TYPE_CHECKING:
    from google import genai


@env.register_client('gemini')
def _(api_key: str) -> 'genai.client.Client':
    from google import genai

    return genai.client.Client(api_key=api_key)


def _genai_client() -> 'genai.client.Client':
    return env.Env.get().get_client('gemini')


@pxt.udf(resource_pool='request-rate:gemini')
async def generate_content(
    contents: str, *, model: str, config: Optional[dict] = None, tools: Optional[list[dict]] = None
) -> dict:
    """
    Generate content from the specified model. For additional details, see:
    <https://ai.google.dev/gemini-api/docs/text-generation>

    Request throttling:
    Applies the rate limit set in the config (section `gemini`, key `rate_limit`). If no rate
    limit is configured, uses a default of 600 RPM.

    __Requirements:__

    - `pip install google-genai`

    Args:
        contents: The input content to generate from.
        model: The name of the model to use.
        config: Configuration for generation, corresponding to keyword arguments of
            `genai.types.GenerateContentConfig`. For details on the parameters, see:
            <https://googleapis.github.io/python-genai/genai.html#module-genai.types>
        tools: An optional list of Pixeltable tools to use. It is also possible to specify tools manually via the
            `config['tools']` parameter, but at most one of `config['tools']` or `tools` may be used.

    Returns:
        A dictionary containing the response and other metadata.

    Examples:
        Add a computed column that applies the model `gemini-2.0-flash`
        to an existing Pixeltable column `tbl.prompt` of the table `tbl`:

        >>> tbl.add_computed_column(response=generate_content(tbl.prompt, model='gemini-2.0-flash'))
    """
    env.Env.get().require_package('google.genai')
    from google.genai import types

    config_: types.GenerateContentConfig
    if config is None and tools is None:
        config_ = None
    else:
        if config is None:
            config_ = types.GenerateContentConfig()
        else:
            config_ = types.GenerateContentConfig(**config)
        if tools is not None:
            gemini_tools = [__convert_pxt_tool(tool) for tool in tools]
            config_.tools = [types.Tool(function_declarations=gemini_tools)]

    response = await _genai_client().aio.models.generate_content(model=model, contents=contents, config=config_)
    return response.model_dump()


def __convert_pxt_tool(pxt_tool: dict) -> dict:
    return {
        'name': pxt_tool['name'],
        'description': pxt_tool['description'],
        'parameters': {
            'type': 'object',
            'properties': pxt_tool['parameters']['properties'],
            'required': pxt_tool['required'],
        },
    }


def invoke_tools(tools: pxt.func.Tools, response: exprs.Expr) -> exprs.InlineDict:
    """Converts an OpenAI response dict to Pixeltable tool invocation format and calls `tools._invoke()`."""
    return tools._invoke(_gemini_response_to_pxt_tool_calls(response))


@pxt.udf
def _gemini_response_to_pxt_tool_calls(response: dict) -> Optional[dict]:
    pxt_tool_calls: dict[str, list[dict]] = {}
    for part in response['candidates'][0]['content']['parts']:
        tool_call = part.get('function_call')
        if tool_call is not None:
            tool_name = tool_call['name']
            if tool_name not in pxt_tool_calls:
                pxt_tool_calls[tool_name] = []
            pxt_tool_calls[tool_name].append({'args': tool_call['args']})
    if len(pxt_tool_calls) == 0:
        return None
    return pxt_tool_calls


@generate_content.resource_pool
def _(model: str) -> str:
    return f'request-rate:gemini:{model}'


@pxt.udf(resource_pool='request-rate:imagen')
async def generate_images(prompt: str, *, model: str, config: Optional[dict] = None) -> PIL.Image.Image:
    """
    Generates images based on a text description and configuration. For additional details, see:
    <https://ai.google.dev/gemini-api/docs/image-generation>

    __Requirements:__

    - `pip install google-genai`

    Args:
        prompt: A text description of the images to generate.
        model: The model to use.
        config: Configuration for generation, corresponding to keyword arguments of
            `genai.types.GenerateImagesConfig`. For details on the parameters, see:
            <https://googleapis.github.io/python-genai/genai.html#module-genai.types>

    Returns:
        The generated image.

    Examples:
        Add a computed column that applies the model `imagen-3.0-generate-002`
        to an existing Pixeltable column `tbl.prompt` of the table `tbl`:

        >>> tbl.add_computed_column(response=generate_images(tbl.prompt, model='imagen-3.0-generate-002'))
    """
    env.Env.get().require_package('google.genai')
    from google.genai.types import GenerateImagesConfig

    config_ = GenerateImagesConfig(**config) if config else None
    response = await _genai_client().aio.models.generate_images(model=model, prompt=prompt, config=config_)
    return response.generated_images[0].image._pil_image


@generate_images.resource_pool
def _(model: str) -> str:
    return f'request-rate:imagen:{model}'


@pxt.udf(resource_pool='request-rate:veo')
async def generate_videos(
    prompt: Optional[str] = None, image: Optional[PIL.Image.Image] = None, *, model: str, config: Optional[dict] = None
) -> pxt.Video:
    """
    Generates videos based on a text description and configuration. For additional details, see:
    <https://ai.google.dev/gemini-api/docs/video-generation>

    __Requirements:__

    - `pip install google-genai`

    Args:
        prompt: A text description of the videos to generate.
        image: An optional image to use as the first frame of the video. At least one of `prompt` or `image` must be
            provided. (It is ok to specify both.)
        model: The model to use.
        config: Configuration for generation, corresponding to keyword arguments of
            `genai.types.GenerateVideosConfig`. For details on the parameters, see:
            <https://googleapis.github.io/python-genai/genai.html#module-genai.types>

    Returns:
        The generated video.

    Examples:
        Add a computed column that applies the model `veo-2.0-generate-001`
        to an existing Pixeltable column `tbl.prompt` of the table `tbl`:

        >>> tbl.add_computed_column(response=generate_videos(tbl.prompt, model='veo-2.0-generate-001'))
    """
    env.Env.get().require_package('google.genai')
    from google.genai import types

    if prompt is None and image is None:
        raise excs.Error('At least one of `prompt` or `image` must be provided.')

    image_: Optional[types.Image] = None
    if image is not None:
        with io.BytesIO() as buffer:
            image.save(buffer, format='jpeg')
            image_ = types.Image(image_bytes=buffer.getvalue(), mime_type='image/jpeg')

    config_ = types.GenerateVideosConfig(**config) if config else None
    operation = await _genai_client().aio.models.generate_videos(
        model=model, prompt=prompt, image=image_, config=config_
    )
    while not operation.done:
        await asyncio.sleep(3)
        operation = await _genai_client().aio.operations.get(operation)

    video = operation.response.generated_videos[0]

    video_bytes = await _genai_client().aio.files.download(file=video.video)  # type: ignore[arg-type]
    assert video_bytes is not None

    # Create a temporary file to store the video bytes
    output_path = env.Env.get().create_tmp_path('.mp4')
    Path(output_path).write_bytes(video_bytes)
    return str(output_path)


@generate_videos.resource_pool
def _(model: str) -> str:
    return f'request-rate:veo:{model}'



from typing import Any, Optional

import sqlalchemy as sql

from pixeltable.metadata import register_converter
from pixeltable.metadata.converters.util import convert_table_md


@register_converter(version=38)
def _(engine: sql.engine.Engine) -> None:
    convert_table_md(engine, substitution_fn=__substitute_md)


def __substitute_md(k: Optional[str], v: Any) -> Optional[tuple[Optional[str], Any]]:
    if k == 'col_mapping':
        assert isinstance(v, list)
        return k, [__col_mapping_entry(e) for e in v]
    if k == 'stored_proxies':
        assert isinstance(v, list)
        return k, [__stored_proxies_entry(e) for e in v]
    return None


def __col_mapping_entry(e: list) -> list:
    assert isinstance(e, list)
    assert isinstance(e[0], dict)
    assert isinstance(e[1], str)
    return [__col_handle(e[0]), e[1]]


def __stored_proxies_entry(e: list) -> list:
    assert isinstance(e, list)
    assert isinstance(e[0], dict)
    assert isinstance(e[1], dict)
    return [__col_handle(e[0]), __col_handle(e[1])]


def __col_handle(e: dict) -> dict:
    return {'tbl_version': {'id': e['tbl_id'], 'effective_version': None}, 'col_id': e['col_id']}



from types import TracebackType
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from pixeltable import exprs


class Error(Exception):
    pass


class ExprEvalError(Exception):
    """
    Used during query execution to signal expr evaluation failures.

    NOT A USER-FACING EXCEPTION. All ExprEvalError instances need to be converted into Error instances.
    """

    expr: 'exprs.Expr'
    expr_msg: str
    exc: Exception
    exc_tb: TracebackType
    input_vals: list[Any]
    row_num: int

    def __init__(
        self,
        expr: 'exprs.Expr',
        expr_msg: str,
        exc: Exception,
        exc_tb: TracebackType,
        input_vals: list[Any],
        row_num: int,
    ) -> None:
        exct = type(exc)
        super().__init__(
            f'Expression evaluation failed with an error of type `{exct.__module__}.{exct.__qualname__}`:\n{expr}'
        )
        self.expr = expr
        self.expr_msg = expr_msg
        self.exc = exc
        self.exc_tb = exc_tb
        self.input_vals = input_vals
        self.row_num = row_num


class PixeltableWarning(Warning):
    pass



import sqlalchemy as sql

from pixeltable.metadata import register_converter


@register_converter(version=35)
def _(engine: sql.engine.Engine) -> None:
    with engine.begin() as conn:
        conn.execute(sql.text('ALTER TABLE tables ADD COLUMN lock_dummy int8'))



from typing import Any

import sqlalchemy as sql

from pixeltable.metadata import register_converter
from pixeltable.metadata.schema import Table


@register_converter(version=14)
def _(engine: sql.engine.Engine) -> None:
    default_remotes: dict[str, Any] = {'remotes': []}
    with engine.begin() as conn:
        conn.execute(sql.update(Table).where(Table.md['remotes'] == None).values(md=Table.md.concat(default_remotes)))



"""
Pixeltable [UDFs](https://pixeltable.readme.io/docs/user-defined-functions-udfs) for `DateType`.

Usage example:
```python
import pixeltable as pxt

t = pxt.get_table(...)
t.select(t.date_col.year, t.date_col.weekday()).collect()
```
"""

from datetime import date, timedelta

import sqlalchemy as sql

import pixeltable as pxt
from pixeltable.utils.code import local_public_names

_SQL_ZERO = sql.literal(0)

# NOT YET SUPPORTED date +/- integer
# NOT YET SUPPORTED date1 - date2 -> integer
# NOT YET SUPPORTED timestamp(date)
# NOT YET SUPPORTED date(timestamp)


@pxt.udf(is_property=True)
def year(self: date) -> int:
    """
    Between [`MINYEAR`](https://docs.python.org/3/library/datetime.html#datetime.MINYEAR) and
    [`MAXYEAR`](https://docs.python.org/3/library/datetime.html#datetime.MAXYEAR) inclusive.

    Equivalent to [`date.year`](https://docs.python.org/3/library/datetime.html#datetime.date.year).
    """
    return self.year


@year.to_sql
def _(self: sql.ColumnElement) -> sql.ColumnElement:
    return sql.extract('year', self)


@pxt.udf(is_property=True)
def month(self: date) -> int:
    """
    Between 1 and 12 inclusive.

    Equivalent to [`date.month`](https://docs.python.org/3/library/datetime.html#datetime.date.month).
    """
    return self.month


@month.to_sql
def _(self: sql.ColumnElement) -> sql.ColumnElement:
    return sql.extract('month', self)


@pxt.udf(is_property=True)
def day(self: date) -> int:
    """
    Between 1 and the number of days in the given month of the given year.

    Equivalent to [`date.day`](https://docs.python.org/3/library/datetime.html#datetime.date.day).
    """
    return self.day


@day.to_sql
def _(self: sql.ColumnElement) -> sql.ColumnElement:
    return sql.extract('day', self)


@pxt.udf(is_method=True)
def make_date(year: int, month: int, day: int) -> date:
    """
    Create a date.

    Equivalent to [`datetime()`](https://docs.python.org/3/library/datetime.html#datetime.date).
    """
    return date(year, month, day)


@make_date.to_sql
def _(year: sql.ColumnElement, month: sql.ColumnElement, day: sql.ColumnElement) -> sql.ColumnElement:
    return sql.func.make_date(sql.cast(year, sql.Integer), sql.cast(month, sql.Integer), sql.cast(day, sql.Integer))


@pxt.udf(is_method=True)
def weekday(self: date) -> int:
    """
    Between 0 (Monday) and 6 (Sunday) inclusive.

    Equivalent to [`date.weekday()`](https://docs.python.org/3/library/datetime.html#datetime.date.weekday).
    """
    return self.weekday()


@weekday.to_sql
def _(self: sql.ColumnElement) -> sql.ColumnElement:
    return sql.extract('isodow', self) - 1


@pxt.udf(is_method=True)
def isoweekday(self: date) -> int:
    """
    Return the day of the week as an integer, where Monday is 1 and Sunday is 7.

    Equivalent to [`date.isoweekday()`](https://docs.python.org/3/library/datetime.html#datetime.date.isoweekday).
    """
    return self.isoweekday()


@isoweekday.to_sql
def _(self: sql.ColumnElement) -> sql.ColumnElement:
    return sql.extract('isodow', self)


@pxt.udf(is_method=True)
def isocalendar(self: date) -> dict:
    """
    Return a dictionary with three entries: `'year'`, `'week'`, and `'weekday'`.

    Equivalent to
    [`date.isocalendar()`](https://docs.python.org/3/library/datetime.html#datetime.date.isocalendar).
    """
    iso_year, iso_week, iso_weekday = self.isocalendar()
    return {'year': iso_year, 'week': iso_week, 'weekday': iso_weekday}


@pxt.udf(is_method=True)
def isoformat(self: date, sep: str = 'T', timespec: str = 'auto') -> str:
    """
    Return a string representing the date and time in ISO 8601 format.

    Equivalent to [`date.isoformat()`](https://docs.python.org/3/library/datetime.html#datetime.date.isoformat).

    Args:
        sep: Separator between date and time.
        timespec: The number of additional terms in the output. See the
            [`date.isoformat()`](https://docs.python.org/3/library/datetime.html#datetime.date.isoformat)
            documentation for more details.
    """
    return self.isoformat()


@pxt.udf(is_method=True)
def toordinal(self: date) -> int:
    """
    Return the proleptic Gregorian ordinal of the date, where January 1 of year 1 has ordinal 1.

    Equivalent to [`date.toordinal()`](https://docs.python.org/3/library/datetime.html#datetime.date.toordinal).
    """
    return self.toordinal()


@pxt.udf(is_method=True)
def strftime(self: date, format: str) -> str:
    """
    Return a string representing the date and time, controlled by an explicit format string.

    Equivalent to [`date.strftime()`](https://docs.python.org/3/library/datetime.html#datetime.date.strftime).

    Args:
        format: The format string to control the output. For a complete list of formatting directives, see
            [`strftime()` and `strptime()` Behavior](https://docs.python.org/3/library/datetime.html#strftime-and-strptime-behavior).
    """
    return self.strftime(format)


@pxt.udf(is_method=True)
def add_days(self: date, n: int) -> date:
    """
    Add `n` days to the date.

    Equivalent to [`date + timedelta(days=n)`](https://docs.python.org/3/library/datetime.html#datetime.timedelta).
    """
    return self + timedelta(days=n)


__all__ = local_public_names(__name__)


def __dir__() -> list[str]:
    return __all__



import base64
import datetime
import io
import itertools
import json
import logging
import tarfile
import urllib.parse
import urllib.request
import uuid
from pathlib import Path
from typing import Any, Iterator, Optional
from uuid import UUID

import more_itertools
import numpy as np
import PIL.Image
import pyarrow as pa
import pyarrow.parquet as pq
import sqlalchemy as sql

import pixeltable as pxt
from pixeltable import catalog, exceptions as excs, metadata, type_system as ts
from pixeltable.env import Env
from pixeltable.metadata import schema
from pixeltable.utils import sha256sum
from pixeltable.utils.formatter import Formatter
from pixeltable.utils.media_store import MediaStore

_logger = logging.getLogger('pixeltable')


class TablePackager:
    """
    Packages a pixeltable Table into a tarball containing Parquet tables and media files. The structure of the tarball
    is as follows:

    metadata.json  # Pixeltable metadata for the packaged table and its ancestors
    tables/**  # Parquet tables for the packaged table and its ancestors, each table in a directory 'tbl_{tbl_id.hex}'
    media/**  # Local media files

    If the table contains media columns, they are handled as follows:
    - If a media file has an external URL (any URL scheme other than file://), then the URL will be preserved as-is and
      stored in the Parquet table.
    - If a media file is a local file, then it will be copied into the tarball as a file of the form
      'media/{uuid}{extension}', and the Parquet table will contain the ephemeral URI 'pxtmedia://{uuid}{extension}'.
    """

    table: catalog.Table  # The table to be packaged
    tmp_dir: Path  # Temporary directory where the package will reside
    tables_dir: Path  # Directory where the Parquet tables will be written
    media_files: dict[Path, str]  # Mapping from local media file paths to their tarball names
    md: dict[str, Any]

    bundle_path: Path
    preview_header: dict[str, str]
    preview: list[list[Any]]

    def __init__(self, table: catalog.Table, additional_md: Optional[dict[str, Any]] = None) -> None:
        self.table = table
        self.tmp_dir = Path(Env.get().create_tmp_path())
        self.media_files = {}

        # Load metadata
        with catalog.Catalog.get().begin_xact(for_write=False):
            tbl_md = catalog.Catalog.get().load_replica_md(table)
            self.md = {
                'pxt_version': pxt.__version__,
                'pxt_md_version': metadata.VERSION,
                'md': {'tables': [md.as_dict() for md in tbl_md]},
            }
        if additional_md is not None:
            self.md.update(additional_md)

    def package(self) -> Path:
        """
        Export the table to a tarball containing Parquet tables and media files.
        """
        assert not self.tmp_dir.exists()  # Packaging can only be done once per TablePackager instance

        _logger.info(f'Packaging table {self.table._path()!r} and its ancestors in: {self.tmp_dir}')
        self.tmp_dir.mkdir()
        with open(self.tmp_dir / 'metadata.json', 'w', encoding='utf8') as fp:
            json.dump(self.md, fp)
        self.tables_dir = self.tmp_dir / 'tables'
        self.tables_dir.mkdir()
        with catalog.Catalog.get().begin_xact(for_write=False):
            for tv in self.table._tbl_version_path.get_tbl_versions():
                _logger.info(f'Exporting table {tv.get().versioned_name!r}.')
                self.__export_table(tv.get())

        _logger.info('Building archive.')
        self.bundle_path = self.__build_tarball()

        _logger.info('Extracting preview data.')
        self.md['count'] = self.table.count()
        preview_header, preview = self.__extract_preview_data()
        self.md['preview_header'] = preview_header
        self.md['preview'] = preview

        _logger.info(f'Packaging complete: {self.bundle_path}')
        return self.bundle_path

    def __export_table(self, tv: catalog.TableVersion) -> None:
        """
        Exports the data from `t` into a Parquet table.
        """
        # `tv` must be an ancestor of the primary table
        assert any(tv.id == base.id for base in self.table._tbl_version_path.get_tbl_versions())
        sql_types = {col.name: col.type for col in tv.store_tbl.sa_tbl.columns}
        media_cols: set[str] = set()
        for col in tv.cols:
            if col.is_stored and col.col_type.is_media_type():
                media_cols.add(col.store_name())

        parquet_schema = self.__to_parquet_schema(tv.store_tbl.sa_tbl)
        # TODO: Partition larger tables into multiple parquet files. (The parquet file naming scheme anticipates
        #     future support for this.)
        parquet_dir = self.tables_dir / f'tbl_{tv.id.hex}'
        parquet_dir.mkdir()
        parquet_file = parquet_dir / f'tbl_{tv.id.hex}.00000.parquet'
        _logger.info(f'Creating parquet table: {parquet_file}')

        # Populate the Parquet table with data.
        # The data is first loaded from the DataFrame into a sequence of pyarrow tables, batched in order to avoid
        # excessive memory usage. The pyarrow tables are then amalgamated into the (single) Parquet table on disk.
        # We use snappy compression for the Parquet tables; the entire bundle will be bzip2-compressed later, so
        # faster compression should provide good performance while still reducing temporary storage utilization.
        parquet_writer = pq.ParquetWriter(parquet_file, parquet_schema, compression='SNAPPY')
        filter_tv = self.table._tbl_version_path.tbl_version.get()
        row_iter = tv.store_tbl.dump_rows(tv.version, filter_tv.store_tbl, filter_tv.version)
        for pa_table in self.__to_pa_tables(row_iter, sql_types, media_cols, parquet_schema):
            parquet_writer.write_table(pa_table)
        parquet_writer.close()

    # The following methods are responsible for schema and data conversion from Pixeltable to Parquet.

    @classmethod
    def __to_parquet_schema(cls, store_tbl: sql.Table) -> pa.Schema:
        entries = [(col_name, cls.__to_parquet_type(col.type)) for col_name, col in store_tbl.columns.items()]
        return pa.schema(entries)  # type: ignore[arg-type]

    @classmethod
    def __to_parquet_type(cls, col_type: sql.types.TypeEngine[Any]) -> pa.DataType:
        if isinstance(col_type, sql.String):
            return pa.string()
        if isinstance(col_type, sql.Boolean):
            return pa.bool_()
        if isinstance(col_type, sql.BigInteger):
            return pa.int64()
        if isinstance(col_type, sql.Float):
            return pa.float32()
        if isinstance(col_type, sql.TIMESTAMP):
            return pa.timestamp('us', tz=datetime.timezone.utc)
        if isinstance(col_type, sql.Date):
            return pa.date32()
        if isinstance(col_type, sql.JSON):
            return pa.string()  # JSON will be exported as strings
        if isinstance(col_type, sql.LargeBinary):
            return pa.binary()
        raise AssertionError(f'Unrecognized SQL type: {col_type} (type {type(col_type)})')

    def __to_pa_tables(
        self,
        row_iter: Iterator[dict[str, Any]],
        sql_types: dict[str, sql.types.TypeEngine[Any]],
        media_cols: set[str],
        arrow_schema: pa.Schema,
        batch_size: int = 1_000,
    ) -> Iterator[pa.Table]:
        """
        Group rows into a sequence of pyarrow tables, batched into smaller chunks to minimize memory utilization.
        The row dictionaries have the format {store_col_name: value}, where the values reflect the unprocessed contents
        of the store database (as returned by `StoreTable.dump_rows()`).
        """
        for rows in more_itertools.batched(row_iter, batch_size):
            cols = {}
            for name, sql_type in sql_types.items():
                is_media_col = name in media_cols
                values = [self.__to_pa_value(row.get(name), sql_type, is_media_col) for row in rows]
                cols[name] = values
            yield pa.Table.from_pydict(cols, schema=arrow_schema)

    def __to_pa_value(self, val: Any, sql_type: sql.types.TypeEngine[Any], is_media_col: bool) -> Any:
        if val is None:
            return None
        if isinstance(sql_type, sql.JSON):
            # Export JSON as strings
            return json.dumps(val)
        if is_media_col:
            # Handle media files as described above
            assert isinstance(val, str)
            return self.__process_media_url(val)
        return val

    def __process_media_url(self, url: str) -> str:
        parsed_url = urllib.parse.urlparse(url)
        if parsed_url.scheme == 'file':
            # It's the URL of a local file. Replace it with a pxtmedia:// URI.
            # (We can't use an actual pxt:// URI, because the eventual pxt:// table name might not be known at this
            # time. The pxtmedia:// URI serves as a relative reference into the tarball that can be replaced with an
            # actual URL when the table is reconstituted.)
            path = Path(urllib.parse.unquote(urllib.request.url2pathname(parsed_url.path)))
            if path not in self.media_files:
                # Create a new entry in the `media_files` dict so that we can copy the file into the tarball later.
                # We name the media files in the archive by their SHA256 hash. This ensures that we can properly
                # deduplicate and validate them later.
                # If we get a collision, it's not a problem; it just means we have two identical files (which will
                # be conveniently deduplicated in the bundle).
                sha = sha256sum(path)
                dest_name = f'{sha}{path.suffix}'
                self.media_files[path] = dest_name
            return f'pxtmedia://{self.media_files[path]}'
        # For any type of URL other than a local file, just return the URL as-is.
        return url

    def __build_tarball(self) -> Path:
        bundle_path = self.tmp_dir / 'bundle.tar.bz2'
        with tarfile.open(bundle_path, 'w:bz2') as tf:
            # Add metadata json
            tf.add(self.tmp_dir / 'metadata.json', arcname='metadata.json')
            # Add the dir containing Parquet tables
            tf.add(self.tables_dir, arcname='tables')
            # Add the media files
            for src_file, dest_name in self.media_files.items():
                tf.add(src_file, arcname=f'media/{dest_name}')
        return bundle_path

    def __extract_preview_data(self) -> tuple[dict[str, str], list[list[Any]]]:
        """
        Extract a preview of the table data for display in the UI.

        In order to bound the size of the output data, all "unbounded" data types are resized:
        - Strings are abbreviated as per Formatter.abbreviate()
        - Arrays and JSON are shortened and formatted as strings
        - Images are resized to thumbnail size as a base64-encoded webp
        - Videos are replaced by their first frame and resized as above
        - Documents are replaced by a thumbnail as a base64-encoded webp
        """
        # First 8 columns
        preview_cols = dict(itertools.islice(self.table._get_schema().items(), 0, 8))
        select_list = [self.table[col_name] for col_name in preview_cols]
        # First 5 rows
        rows = list(self.table.select(*select_list).head(n=5))

        preview_header = {col_name: str(col_type._type) for col_name, col_type in preview_cols.items()}
        preview = [
            [self.__encode_preview_data(val, col_type)]
            for row in rows
            for val, col_type in zip(row.values(), preview_cols.values())
        ]

        return preview_header, preview

    def __encode_preview_data(self, val: Any, col_type: ts.ColumnType) -> Any:
        if val is None:
            return None

        match col_type._type:
            case ts.ColumnType.Type.STRING:
                assert isinstance(val, str)
                return Formatter.abbreviate(val)

            case ts.ColumnType.Type.INT | ts.ColumnType.Type.FLOAT | ts.ColumnType.Type.BOOL:
                return val

            case ts.ColumnType.Type.TIMESTAMP | ts.ColumnType.Type.DATE:
                return str(val)

            case ts.ColumnType.Type.ARRAY:
                assert isinstance(val, np.ndarray)
                return Formatter.format_array(val)

            case ts.ColumnType.Type.JSON:
                # We need to escape the JSON string server-side for security reasons.
                # Therefore we don't escape it here, in order to avoid double-escaping.
                return Formatter.format_json(val, escape_strings=False)

            case ts.ColumnType.Type.IMAGE:
                # Rescale the image to minimize data transfer size
                assert isinstance(val, PIL.Image.Image)
                return self.__encode_image(val)

            case ts.ColumnType.Type.VIDEO:
                assert isinstance(val, str)
                return self.__encode_video(val)

            case ts.ColumnType.Type.AUDIO:
                return None

            case ts.ColumnType.Type.DOCUMENT:
                assert isinstance(val, str)
                return self.__encode_document(val)

            case _:
                raise AssertionError(f'Unrecognized column type: {col_type._type}')

    def __encode_image(self, img: PIL.Image.Image) -> str:
        # Heuristic for thumbnail sizing:
        # Standardize on a width of 240 pixels (to most efficiently utilize the columnar display).
        # But, if the aspect ratio is below 2:3, bound the height at 360 pixels (to avoid unboundedly tall thumbnails
        #     in the case of highly oblong images).
        if img.height > img.width * 1.5:
            scaled_img = img.resize((img.width * 360 // img.height, 360))
        else:
            scaled_img = img.resize((240, img.height * 240 // img.width))
        with io.BytesIO() as buffer:
            scaled_img.save(buffer, 'webp')
            return base64.b64encode(buffer.getvalue()).decode()

    def __encode_video(self, video_path: str) -> Optional[str]:
        thumb = Formatter.extract_first_video_frame(video_path)
        return self.__encode_image(thumb) if thumb is not None else None

    def __encode_document(self, doc_path: str) -> Optional[str]:
        thumb = Formatter.make_document_thumbnail(doc_path)
        return self.__encode_image(thumb) if thumb is not None else None


class TableRestorer:
    """
    Creates a replica table from a tarball containing Parquet tables and media files. See the `TablePackager` docs for
    details on the tarball structure.

    Args:
        tbl_path: Pixeltable path (such as 'my_dir.my_table') where the materialized table will be made visible.
        md: Optional metadata dictionary. If not provided, metadata will be read from the tarball's `metadata.json`.
            The metadata contains table_md, table_version_md, and table_schema_version_md entries for each ancestor
            of the table being restored, as written out by `TablePackager`.
    """

    tbl_path: str
    md: Optional[dict[str, Any]]
    tmp_dir: Path
    media_files: dict[str, str]  # Mapping from pxtmedia:// URLs to local file:// URLs

    def __init__(self, tbl_path: str, md: Optional[dict[str, Any]] = None) -> None:
        self.tbl_path = tbl_path
        self.md = md
        self.tmp_dir = Path(Env.get().create_tmp_path())
        self.media_files = {}

    def restore(self, bundle_path: Path) -> pxt.Table:
        # Extract tarball
        print(f'Extracting table data into: {self.tmp_dir}')
        with tarfile.open(bundle_path, 'r:bz2') as tf:
            tf.extractall(path=self.tmp_dir)

        if self.md is None:
            # No metadata supplied; read it from the archive
            with open(self.tmp_dir / 'metadata.json', 'r', encoding='utf8') as fp:
                self.md = json.load(fp)

        pxt_md_version = self.md['pxt_md_version']
        assert isinstance(pxt_md_version, int)

        if pxt_md_version != metadata.VERSION:
            raise excs.Error(
                f'Pixeltable metadata version mismatch: {pxt_md_version} != {metadata.VERSION}.\n'
                'Please upgrade Pixeltable to use this dataset: pip install -U pixeltable'
            )

        tbl_md = [schema.FullTableMd.from_dict(t) for t in self.md['md']['tables']]
        for md in tbl_md:
            md.tbl_md.is_replica = True

        cat = catalog.Catalog.get()

        with cat.begin_xact(for_write=True):
            # Create (or update) the replica table and its ancestors, along with TableVersion instances for any
            # versions that have not been seen before.
            cat.create_replica(catalog.Path(self.tbl_path), tbl_md)

            # Now we need to load data for replica_tbl and its ancestors, except that we skip
            # replica_tbl itself if it's a pure snapshot.
            for md in tbl_md[::-1]:  # Base table first
                if not md.is_pure_snapshot:
                    tv = cat.get_tbl_version(UUID(md.tbl_md.tbl_id), md.version_md.version)
                    # Import data from Parquet.
                    _logger.info(f'Importing table {tv.name!r}.')
                    self.__import_table(self.tmp_dir, tv, md)

            return cat.get_table_by_id(UUID(tbl_md[0].tbl_md.tbl_id))

    def __import_table(self, bundle_path: Path, tv: catalog.TableVersion, tbl_md: schema.FullTableMd) -> None:
        """
        Import the Parquet table into the Pixeltable catalog.
        """
        tbl_id = UUID(tbl_md.tbl_md.tbl_id)
        parquet_dir = bundle_path / 'tables' / f'tbl_{tbl_id.hex}'
        parquet_table = pq.read_table(str(parquet_dir))
        replica_version = tv.version

        conn = Env.get().conn
        store_sa_tbl = tv.store_tbl.sa_tbl
        store_sa_tbl_name = tv.store_tbl._storage_name()

        # Sometimes we are importing a table that has never been seen before. Other times, however, we are importing
        # an existing replica table, and the table version and/or row selection differs from what was imported
        # previously. Care must be taken to ensure that the new data is merged with existing data in a way that
        # yields an internally consistent version history for each row.

        # The overall strategy is this:
        # 1. Import the parquet data into a temporary table;
        # 2. "rectify" the v_max values in both the temporary table and the existing table (more on this below);
        # 3. Delete any row instances from the temporary table that are already present in the existing table;
        # 4. Copy the remaining rows from the temporary table into the existing table.

        # Create a temporary table for the initial data load, containing columns for all columns present in the
        # parquet table. The parquet columns have identical names to those in the store table, so we can use the
        # store table schema to get their SQL types (which are not necessarily derivable from their Parquet types,
        # e.g., pa.string() may hold either VARCHAR or serialized JSONB).
        temp_cols: dict[str, sql.Column] = {}
        for field in parquet_table.schema:
            assert field.name in store_sa_tbl.columns
            col_type = store_sa_tbl.columns[field.name].type
            temp_cols[field.name] = sql.Column(field.name, col_type)
        temp_sa_tbl_name = f'temp_{uuid.uuid4().hex}'
        _logger.debug(f'Creating temporary table: {temp_sa_tbl_name}')
        temp_md = sql.MetaData()
        temp_sa_tbl = sql.Table(temp_sa_tbl_name, temp_md, *temp_cols.values(), prefixes=['TEMPORARY'])
        temp_sa_tbl.create(conn)

        # Populate the temporary table with data from the Parquet file.
        _logger.debug(f'Loading {parquet_table.num_rows} row(s) into temporary table: {temp_sa_tbl_name}')
        for batch in parquet_table.to_batches(max_chunksize=10_000):
            pydict = batch.to_pydict()
            rows = self.__from_pa_pydict(tv, pydict)
            conn.execute(sql.insert(temp_sa_tbl), rows)

        # Each row version is identified uniquely by its pk, a tuple (row_id, pos_0, pos_1, ..., pos_k, v_min).
        # Conversely, v_max is not part of the primary key, but is simply a bookkeeping device.
        # In an original table, v_max is always equal to the v_min of the succeeding row instance with the same
        # row id, or MAX_VERSION if no such row instance exists. But in the replica, we need to be careful, since
        # we might see only a subset of the original table's versions, and we might see them out of order.

        # We'll adjust the v_max values according to the principle of "latest provable v_max":
        # they will always correspond to the latest version for which we can prove the row instance was alive. This
        # will enable us to maintain consistency of the v_max values if additional table versions are later imported,
        # regardless of the order in which they are seen. It also means that replica tables (unlike original tables)
        # may have gaps in their row version histories, but this is fine; the gaps simply correspond to table versions
        # that have never been observed.

        pk_predicates = [col == temp_cols[col.name] for col in tv.store_tbl.pk_columns()]
        pk_clause = sql.and_(*pk_predicates)

        # If the same pk exists in both the temporary table and the existing table, then the corresponding row data
        # must be identical; the rows can differ only in their v_max value. As a sanity check, we go through the
        # motion of verifying this; a failure implies data corruption in either the replica being imported or in a
        # previously imported replica.

        system_col_names = {col.name for col in tv.store_tbl.system_columns()}
        media_col_names = {col.store_name() for col in tv.cols if col.col_type.is_media_type() and col.is_stored}
        value_store_cols = [
            store_sa_tbl.c[col_name]
            for col_name in temp_cols
            if col_name not in system_col_names and col_name not in media_col_names
        ]
        value_temp_cols = [
            col
            for col_name, col in temp_cols.items()
            if col_name not in system_col_names and col_name not in media_col_names
        ]
        mismatch_predicates = [store_col != temp_col for store_col, temp_col in zip(value_store_cols, value_temp_cols)]
        mismatch_clause = sql.or_(*mismatch_predicates)

        # This query looks for rows that have matching primary keys (rowid + pos_k + v_min), but differ in at least
        # one value column. Pseudo-SQL:
        #
        # SELECT store_tbl.col_0, ..., store_tbl.col_n, temp_tbl.col_0, ...,  temp_tbl.col_n
        # FROM store_tbl, temp_tbl
        # WHERE store_tbl.rowid = temp_tbl.rowid
        #     AND store_tbl.pos_0 = temp_tbl.pos_0
        #     AND ... AND store_tbl.pos_k = temp_tbl.pos_k
        #     AND store_tbl.v_min = temp_tbl.v_min
        #     AND (
        #         store_tbl.col_0 != temp_tbl.col_0
        #         OR store_tbl.col_1 != temp_tbl.col_1
        #         OR ... OR store_tbl.col_n != temp_tbl.col_n
        #     )
        #
        # The value column comparisons (store_tbl.col_0 != temp_tbl.col_0, etc.) will always be false for rows where
        # either column is NULL; this is what we want, since it may indicate a column that is present in one version
        # but not the other.
        q = sql.select(*value_store_cols, *value_temp_cols).where(pk_clause).where(mismatch_clause)
        _logger.debug(q.compile())
        result = conn.execute(q)
        if result.rowcount > 0:
            _logger.debug(
                f'Data corruption error between {temp_sa_tbl_name!r} and {store_sa_tbl_name!r}: '
                f'{result.rowcount} inconsistent row(s).'
            )
            row = result.first()
            _logger.debug('Example mismatch:')
            _logger.debug(f'{store_sa_tbl_name}: {row[: len(value_store_cols)]}')
            _logger.debug(f'{temp_sa_tbl_name}: {row[len(value_store_cols) :]}')
            raise excs.Error(
                'Data corruption error: the replica data are inconsistent with data retrieved from a previous replica.'
            )
        _logger.debug(f'Verified data integrity between {store_sa_tbl_name!r} and {temp_sa_tbl_name!r}.')

        # Now rectify the v_max values in the temporary table.
        # If a row instance has a concrete v_max value, then we know it's genuine: it's the unique and immutable
        # version when the row was deleted. (This can only happen if later versions of the base table already
        # existed at the time this replica was published.)
        # But if a row instance has a v_max value of MAX_VERSION, then we don't know anything about its future.
        # It might live indefinitely, or it might be deleted as early as version `n + 1`. Following the principle
        # of "latest provable v_max", we simply set v_max equal to `n + 1`.
        q = (
            temp_sa_tbl.update()
            .values(v_max=(replica_version + 1))
            .where(temp_sa_tbl.c.v_max == schema.Table.MAX_VERSION)
        )
        _logger.debug(q.compile())
        result = conn.execute(q)
        _logger.debug(f'Rectified {result.rowcount} row(s) in {temp_sa_tbl_name!r}.')

        # Now rectify the v_max values in the existing table. This is done by simply taking the later of the two v_max
        # values (the existing one and the new one) for each row instance, following the "latest provable v_max"
        # principle. Obviously we only need to do this for rows that exist in both tables (it's a simple join).
        q = (
            store_sa_tbl.update()
            .values(v_max=sql.func.greatest(store_sa_tbl.c.v_max, temp_sa_tbl.c.v_max))
            .where(pk_clause)
        )
        _logger.debug(q.compile())
        result = conn.execute(q)
        _logger.debug(f'Rectified {result.rowcount} row(s) in {store_sa_tbl_name!r}.')

        # Now we need to update rows in the existing table that are also present in the temporary table. This is to
        # account for the scenario where the temporary table has columns that are not present in the existing table.
        # (We can't simply replace the rows with their versions in the temporary table, because the converse scenario
        # might also occur; there may be columns in the existing table that are not present in the temporary table.)
        value_update_clauses: dict[str, sql.ColumnElement] = {}
        for temp_col in temp_cols.values():
            if temp_col.name not in system_col_names:
                store_col = store_sa_tbl.c[temp_col.name]
                # Prefer the value from the existing table, substituting the value from the temporary table if it's
                # NULL. This works in all cases (including media columns, where we prefer the existing media file).
                clause = sql.case((store_col == None, temp_col), else_=store_col)
                value_update_clauses[temp_col.name] = clause
        if len(value_update_clauses) > 0:
            q = store_sa_tbl.update().values(**value_update_clauses).where(pk_clause)
            _logger.debug(q.compile())
            result = conn.execute(q)
            _logger.debug(
                f'Merged values from {temp_sa_tbl_name!r} into {store_sa_tbl_name!r} for {result.rowcount} row(s).'
            )

        # Now drop any rows from the temporary table that are also present in the existing table.
        # The v_max values have been rectified, data has been merged into NULL cells, and all other row values have
        # been verified identical.
        # TODO: Delete any media files that were orphaned by this operation (they're necessarily duplicates of media
        #     files that are already present in the existing table).
        q = temp_sa_tbl.delete().where(pk_clause)
        _logger.debug(q.compile())
        result = conn.execute(q)
        _logger.debug(f'Deleted {result.rowcount} row(s) from {temp_sa_tbl_name!r}.')

        # Finally, copy the remaining data (consisting entirely of new row instances) from the temporary table into
        # the actual table.
        q = store_sa_tbl.insert().from_select(
            [store_sa_tbl.c[col_name] for col_name in temp_cols], sql.select(*temp_cols.values())
        )
        _logger.debug(q.compile())
        result = conn.execute(q)
        _logger.debug(f'Inserted {result.rowcount} row(s) from {temp_sa_tbl_name!r} into {store_sa_tbl_name!r}.')

    def __from_pa_pydict(self, tv: catalog.TableVersion, pydict: dict[str, Any]) -> list[dict[str, Any]]:
        # Data conversions from pyarrow to Pixeltable
        sql_types: dict[str, sql.types.TypeEngine[Any]] = {}
        for col_name in pydict:
            assert col_name in tv.store_tbl.sa_tbl.columns
            sql_types[col_name] = tv.store_tbl.sa_tbl.columns[col_name].type
        media_col_ids: dict[str, int] = {}
        for col in tv.cols:
            if col.is_stored and col.col_type.is_media_type():
                media_col_ids[col.store_name()] = col.id

        row_count = len(next(iter(pydict.values())))
        rows: list[dict[str, Any]] = []
        for i in range(row_count):
            row = {
                col_name: self.__from_pa_value(tv, col_vals[i], sql_types[col_name], media_col_ids.get(col_name))
                for col_name, col_vals in pydict.items()
            }
            rows.append(row)

        return rows

    def __from_pa_value(
        self, tv: catalog.TableVersion, val: Any, sql_type: sql.types.TypeEngine[Any], media_col_id: Optional[int]
    ) -> Any:
        if val is None:
            return None
        if isinstance(sql_type, sql.JSON):
            return json.loads(val)
        if media_col_id is not None:
            assert isinstance(val, str)
            return self.__relocate_media_file(tv, media_col_id, val)
        return val

    def __relocate_media_file(self, tv: catalog.TableVersion, media_col_id: int, url: str) -> str:
        # If this is a pxtmedia:// URL, relocate it
        parsed_url = urllib.parse.urlparse(url)
        assert parsed_url.scheme != 'file'  # These should all have been converted to pxtmedia:// URLs
        if parsed_url.scheme == 'pxtmedia':
            if url not in self.media_files:
                # First time seeing this pxtmedia:// URL. Relocate the file to the media store and record the mapping
                # in self.media_files.
                src_path = self.tmp_dir / 'media' / parsed_url.netloc
                # Move the file to the media store and update the URL.
                self.media_files[url] = MediaStore.relocate_local_media_file(src_path, tv.id, media_col_id, tv.version)
            return self.media_files[url]
        # For any type of URL other than a local file, just return the URL as-is.
        return url



from __future__ import annotations

import inspect
from functools import reduce
from typing import TYPE_CHECKING, Any, Callable, Iterable, Optional, Union, overload

from pixeltable import catalog, exceptions as excs, exprs, func, type_system as ts

from .function import Function
from .signature import Signature

if TYPE_CHECKING:
    from pixeltable import DataFrame


class QueryTemplateFunction(Function):
    """A parameterized query/DataFrame from which an executable DataFrame is created with a function call."""

    template_df: Optional['DataFrame']
    self_name: Optional[str]
    _comment: Optional[str]

    @classmethod
    def create(
        cls, template_callable: Callable, param_types: Optional[list[ts.ColumnType]], path: str, name: str
    ) -> QueryTemplateFunction:
        # we need to construct a template df and a signature
        py_sig = inspect.signature(template_callable)
        py_params = list(py_sig.parameters.values())
        params = Signature.create_parameters(py_params=py_params, param_types=param_types)
        # invoke template_callable with parameter expressions to construct a DataFrame with parameters
        var_exprs = [exprs.Variable(param.name, param.col_type) for param in params]
        template_df = template_callable(*var_exprs)
        from pixeltable import DataFrame

        assert isinstance(template_df, DataFrame)
        # we take params and return json
        sig = Signature(return_type=ts.JsonType(), parameters=params)
        return QueryTemplateFunction(template_df, sig, path=path, name=name, comment=inspect.getdoc(template_callable))

    def __init__(
        self,
        template_df: Optional['DataFrame'],
        sig: Signature,
        path: Optional[str] = None,
        name: Optional[str] = None,
        comment: Optional[str] = None,
    ):
        assert sig is not None
        super().__init__([sig], self_path=path)
        self.self_name = name
        self.template_df = template_df
        self._comment = comment

    def _update_as_overload_resolution(self, signature_idx: int) -> None:
        pass  # only one signature supported for QueryTemplateFunction

    @property
    def is_async(self) -> bool:
        return True

    async def aexec(self, *args: Any, **kwargs: Any) -> Any:
        # assert not self.is_polymorphic
        bound_args = self.signature.py_signature.bind(*args, **kwargs).arguments
        # apply defaults, otherwise we might have Parameters left over
        bound_args.update(
            {
                param.name: param.default
                for param in self.signature.parameters.values()
                if param.has_default() and param.name not in bound_args
            }
        )
        bound_df = self.template_df.bind(bound_args)
        result = await bound_df._acollect()
        return list(result)

    @property
    def display_name(self) -> str:
        return self.self_name

    @property
    def name(self) -> str:
        return self.self_name

    def comment(self) -> Optional[str]:
        return self._comment

    def _as_dict(self) -> dict:
        return {'name': self.name, 'signature': self.signature.as_dict(), 'df': self.template_df.as_dict()}

    @classmethod
    def _from_dict(cls, d: dict) -> Function:
        from pixeltable.dataframe import DataFrame

        return cls(DataFrame.from_dict(d['df']), Signature.from_dict(d['signature']), name=d['name'])


@overload
def query(py_fn: Callable) -> QueryTemplateFunction: ...


@overload
def query(*, param_types: Optional[list[ts.ColumnType]] = None) -> Callable[[Callable], QueryTemplateFunction]: ...


def query(*args: Any, **kwargs: Any) -> Any:
    def make_query_template(py_fn: Callable, param_types: Optional[list[ts.ColumnType]]) -> QueryTemplateFunction:
        if py_fn.__module__ != '__main__' and py_fn.__name__.isidentifier():
            # this is a named function in a module
            function_path = f'{py_fn.__module__}.{py_fn.__qualname__}'
        else:
            function_path = None
        query_name = py_fn.__name__
        query_fn = QueryTemplateFunction.create(py_fn, param_types=param_types, path=function_path, name=query_name)
        return query_fn

        # TODO: verify that the inferred return type matches that of the template
        # TODO: verify that the signature doesn't contain batched parameters

    if len(args) == 1:
        assert len(kwargs) == 0 and callable(args[0])
        return make_query_template(args[0], None)
    else:
        assert len(args) == 0 and len(kwargs) == 1 and 'param_types' in kwargs
        return lambda py_fn: make_query_template(py_fn, kwargs['param_types'])


def retrieval_udf(
    table: catalog.Table,
    name: Optional[str] = None,
    description: Optional[str] = None,
    parameters: Optional[Iterable[Union[str, exprs.ColumnRef]]] = None,
    limit: Optional[int] = 10,
) -> func.QueryTemplateFunction:
    """
    Constructs a retrieval UDF for the given table. The retrieval UDF is a UDF whose parameters are
    columns of the table and whose return value is a list of rows from the table. The return value of
    ```python
    f(col1=x, col2=y, ...)
    ```
    will be a list of all rows from the table that match the specified arguments.

    Args:
        table: The table to use as the dataset for the retrieval tool.
        name: The name of the tool. If not specified, then the name of the table will be used by default.
        description: The description of the tool. If not specified, then a default description will be generated.
        parameters: The columns of the table to use as parameters. If not specified, all data columns
            (non-computed columns) will be used as parameters.

            All of the specified parameters will be required parameters of the tool, regardless of their status
            as columns.
        limit: The maximum number of rows to return. If not specified, then all matching rows will be returned.

    Returns:
        A list of dictionaries containing data from the table, one per row that matches the input arguments.
        If there are no matching rows, an empty list will be returned.
    """
    # Argument validation
    col_refs: list[exprs.ColumnRef]
    # TODO: get rid of references to ColumnRef internals and replace instead with a public interface
    col_names = table.columns()
    if parameters is None:
        col_refs = [table[col_name] for col_name in col_names if not table[col_name].col.is_computed]
    else:
        for param in parameters:
            if isinstance(param, str) and param not in col_names:
                raise excs.Error(f'The specified parameter {param!r} is not a column of the table {table._path()!r}')
        col_refs = [table[param] if isinstance(param, str) else param for param in parameters]

    if len(col_refs) == 0:
        raise excs.Error('Parameter list cannot be empty.')

    # Construct the dataframe
    predicates = [col_ref == exprs.Variable(col_ref.col.name, col_ref.col.col_type) for col_ref in col_refs]
    where_clause = reduce(lambda c1, c2: c1 & c2, predicates)
    df = table.select().where(where_clause)
    if limit is not None:
        df = df.limit(limit)

    # Construct the signature
    query_params = [
        func.Parameter(col_ref.col.name, col_ref.col.col_type, inspect.Parameter.POSITIONAL_OR_KEYWORD)
        for col_ref in col_refs
    ]
    query_signature = func.Signature(return_type=ts.JsonType(), parameters=query_params)

    # Construct a name and/or description if not provided
    if name is None:
        name = table._name
    if description is None:
        description = (
            f'Retrieves an entry from the dataset {name!r} that matches the given parameters.\n\nParameters:\n'
        )
        description += '\n'.join(
            [f'    {col_ref.col.name}: of type `{col_ref.col.col_type._to_base_str()}`' for col_ref in col_refs]
        )

    fn = func.QueryTemplateFunction(df, query_signature, name=name, comment=description)
    return fn



from __future__ import annotations

from typing import Any, Optional

import sqlalchemy as sql

import pixeltable.type_system as ts

from .data_row import DataRow
from .expr import Expr, ExprScope
from .json_mapper import JsonMapperDispatch
from .row_builder import RowBuilder
from .sql_element_cache import SqlElementCache


class ObjectRef(Expr):
    """
    Reference to an intermediate result, such as the "scope variable" produced by a JsonMapper.
    The object is generated/materialized elsewhere and establishes a new scope.
    """

    def __init__(self, scope: ExprScope, owner: JsonMapperDispatch):
        # TODO: do we need an Unknown type after all?
        super().__init__(ts.JsonType())  # JsonType: this could be anything
        self._scope = scope
        self.owner = owner
        self.id = self._create_id()

    def _id_attrs(self) -> list[tuple[str, Any]]:
        # We have no components, so we can't rely on the default behavior here (otherwise, all ObjectRef
        # instances will be conflated into a single slot).
        return [('addr', id(self))]

    def substitute(self, subs: dict[Expr, Expr]) -> Expr:
        # Just return self; we need to avoid creating a new id after doing the substitution, because otherwise
        # we'll wind up in a situation where the scope_anchor of the enclosing JsonMapper is different from the
        # nested ObjectRefs inside its target_expr (and therefore occupies a different slot_idx).
        return self

    def scope(self) -> ExprScope:
        return self._scope

    def _equals(self, other: ObjectRef) -> bool:
        return self.id == other.id

    def sql_expr(self, _: SqlElementCache) -> Optional[sql.ColumnElement]:
        return None

    def eval(self, data_row: DataRow, row_builder: RowBuilder) -> None:
        # this will be called, but the value has already been materialized elsewhere
        pass

    def __repr__(self) -> str:
        return f'ObjectRef({self.owner}, {self.id}, {self.owner.id})'



import http
import http.server
import logging
import pathlib
import urllib
from typing import Any

_logger = logging.getLogger('pixeltable.http.server')


def get_file_uri(http_address: str, file_path: str) -> str:
    """Get the URI for a file path, with the given prefix.
    Used in the client to generate a URI
    """
    abs_path = pathlib.Path(file_path)
    assert abs_path.is_absolute()
    url = urllib.request.pathname2url(str(abs_path))
    return f'{http_address}{url}'


class AbsolutePathHandler(http.server.SimpleHTTPRequestHandler):
    """Serves all absolute paths, not just the current directory"""

    def translate_path(self, path: str) -> str:
        """
        Translate a /-separated PATH to the local filename syntax.
        overrides http.server.SimpleHTTPRequestHandler.translate_path

        This is only useful for file serving.

        Code initially taken from there:
        https://github.com/python/cpython/blob/f5406ef454662b98df107775d18ff71ae6849618/Lib/http/server.py#L834
        """
        _logger.info(f'translate path {path=}')
        # abandon query parameters, taken from http.server.SimpleHTTPRequestHandler
        path = path.split('?', 1)[0]
        path = path.split('#', 1)[0]

        path = pathlib.Path(urllib.request.url2pathname(path))
        return str(path)

    def log_message(self, format: str, *args: Any) -> None:
        """override logging to stderr in http.server.BaseHTTPRequestHandler"""
        message = format % args
        _logger.info(message.translate(self._control_char_table))  # type: ignore[attr-defined]


class LoggingHTTPServer(http.server.ThreadingHTTPServer):
    """Avoids polluting stdout and stderr"""

    def handle_error(self, request, client_address) -> None:  # type: ignore[no-untyped-def]
        """override socketserver.TCPServer.handle_error which prints directly to sys.stderr"""
        import traceback

        _logger.error(
            f'Exception occurred during processing of {request=} from {client_address=}\
            \nbacktrace:\n{traceback.format_exc()}\n----\n'
        )


def make_server(address: str, port: int) -> http.server.HTTPServer:
    """Create a file server with pixeltable specific config"""
    return LoggingHTTPServer((address, port), AbsolutePathHandler)


if __name__ == '__main__':
    httpd = make_server('127.0.0.1', 8000)
    print(f'about to server HTTP on {httpd.server_address}')
    httpd.serve_forever()



from typing import TYPE_CHECKING, Optional

import numpy as np

import pixeltable as pxt
from pixeltable import env
from pixeltable.func import Batch
from pixeltable.utils.code import local_public_names

if TYPE_CHECKING:
    import ollama


@env.register_client('ollama')
def _(host: str) -> 'ollama.Client':
    import ollama

    return ollama.Client(host=host)


def _ollama_client() -> Optional['ollama.Client']:
    try:
        return env.Env.get().get_client('ollama')
    except Exception:
        return None


@pxt.udf
def generate(
    prompt: str,
    *,
    model: str,
    suffix: str = '',
    system: str = '',
    template: str = '',
    context: Optional[list[int]] = None,
    raw: bool = False,
    format: Optional[str] = None,
    options: Optional[dict] = None,
) -> dict:
    """
    Generate a response for a given prompt with a provided model.

    Args:
        prompt: The prompt to generate a response for.
        model: The model name.
        suffix: The text after the model response.
        format: The format of the response; must be one of `'json'` or `None`.
        system: System message.
        template: Prompt template to use.
        context: The context parameter returned from a previous call to `generate()`.
        raw: If `True`, no formatting will be applied to the prompt.
        options: Additional options for the Ollama `chat` call, such as `max_tokens`, `temperature`, `top_p`, and
            `top_k`. For details, see the
            [Valid Parameters and Values](https://github.com/ollama/ollama/blob/main/docs/modelfile.md#valid-parameters-and-values)
            section of the Ollama documentation.
    """
    env.Env.get().require_package('ollama')
    import ollama

    client = _ollama_client() or ollama
    return client.generate(
        model=model,
        prompt=prompt,
        suffix=suffix,
        system=system,
        template=template,
        context=context,
        raw=raw,
        format=format,
        options=options,
    ).dict()  # type: ignore[call-overload]


@pxt.udf
def chat(
    messages: list[dict],
    *,
    model: str,
    tools: Optional[list[dict]] = None,
    format: Optional[str] = None,
    options: Optional[dict] = None,
) -> dict:
    """
    Generate the next message in a chat with a provided model.

    Args:
        messages: The messages of the chat.
        model: The model name.
        tools: Tools for the model to use.
        format: The format of the response; must be one of `'json'` or `None`.
        options: Additional options to pass to the `chat` call, such as `max_tokens`, `temperature`, `top_p`, and
            `top_k`. For details, see the
            [Valid Parameters and Values](https://github.com/ollama/ollama/blob/main/docs/modelfile.md#valid-parameters-and-values)
            section of the Ollama documentation.
    """
    env.Env.get().require_package('ollama')
    import ollama

    client = _ollama_client() or ollama
    return client.chat(model=model, messages=messages, tools=tools, format=format, options=options).dict()  # type: ignore[call-overload]


@pxt.udf(batch_size=16)
def embed(
    input: Batch[str], *, model: str, truncate: bool = True, options: Optional[dict] = None
) -> Batch[pxt.Array[(None,), pxt.Float]]:
    """
    Generate embeddings from a model.

    Args:
        input: The input text to generate embeddings for.
        model: The model name.
        truncate: Truncates the end of each input to fit within context length.
            Returns error if false and context length is exceeded.
        options: Additional options to pass to the `embed` call.
            For details, see the
            [Valid Parameters and Values](https://github.com/ollama/ollama/blob/main/docs/modelfile.md#valid-parameters-and-values)
            section of the Ollama documentation.
    """
    env.Env.get().require_package('ollama')
    import ollama

    client = _ollama_client() or ollama
    results = client.embed(model=model, input=input, truncate=truncate, options=options).dict()
    return [np.array(data, dtype=np.float64) for data in results['embeddings']]


__all__ = local_public_names(__name__)


def __dir__() -> list[str]:
    return __all__



from __future__ import annotations

import abc
import logging
from typing import AsyncIterator, Iterable, Iterator, Optional, TypeVar

from pixeltable import exprs
from pixeltable.env import Env

from .data_row_batch import DataRowBatch
from .exec_context import ExecContext

_logger = logging.getLogger('pixeltable')


class ExecNode(abc.ABC):
    """Base class of all execution nodes"""

    output_exprs: Iterable[exprs.Expr]
    row_builder: exprs.RowBuilder
    input: Optional[ExecNode]
    flushed_img_slots: list[int]  # idxs of image slots of our output_exprs dependencies
    stored_img_cols: list[exprs.ColumnSlotIdx]
    ctx: Optional[ExecContext]

    def __init__(
        self,
        row_builder: exprs.RowBuilder,
        output_exprs: Iterable[exprs.Expr],
        input_exprs: Iterable[exprs.Expr],
        input: Optional[ExecNode] = None,
    ):
        assert all(expr.is_valid for expr in output_exprs)
        self.output_exprs = output_exprs
        self.row_builder = row_builder
        self.input = input
        # we flush all image slots that aren't part of our output but are needed to create our output
        output_slot_idxs = {e.slot_idx for e in output_exprs}
        output_dependencies = row_builder.get_dependencies(output_exprs, exclude=input_exprs)
        self.flushed_img_slots = [
            e.slot_idx for e in output_dependencies if e.col_type.is_image_type() and e.slot_idx not in output_slot_idxs
        ]
        self.stored_img_cols = []
        self.ctx = None  # all nodes of a tree share the same context

    def set_ctx(self, ctx: ExecContext) -> None:
        self.ctx = ctx
        if self.input is not None:
            self.input.set_ctx(ctx)

    def set_stored_img_cols(self, stored_img_cols: list[exprs.ColumnSlotIdx]) -> None:
        self.stored_img_cols = stored_img_cols
        # propagate batch size to the source
        if self.input is not None:
            self.input.set_stored_img_cols(stored_img_cols)

    @abc.abstractmethod
    def __aiter__(self) -> AsyncIterator[DataRowBatch]:
        pass

    def __iter__(self) -> Iterator[DataRowBatch]:
        loop = Env.get().event_loop
        aiter = self.__aiter__()
        try:
            while True:
                batch: DataRowBatch = loop.run_until_complete(aiter.__anext__())
                yield batch
        except StopAsyncIteration:
            pass
        # TODO:
        #  - we seem to have some tasks that aren't accounted for by ExprEvalNode and don't get cancelled by the time
        #    we end up here
        # - however, blindly cancelling all pending tasks doesn't work when running in a jupyter environment, which
        #   creates tasks on its own

    def open(self) -> None:
        """Bottom-up initialization of nodes for execution. Must be called before __next__."""
        if self.input is not None:
            self.input.open()
        self._open()

    def close(self) -> None:
        """Frees node resources top-down after execution. Must be called after final __next__."""
        self._close()
        if self.input is not None:
            self.input.close()

    def _open(self) -> None:
        pass

    def _close(self) -> None:
        pass

    T = TypeVar('T', bound='ExecNode')

    def get_node(self, node_class: type[T]) -> Optional[T]:
        if isinstance(self, node_class):
            return self
        if self.input is not None:
            return self.input.get_node(node_class)
        return None

    def set_limit(self, limit: int) -> None:
        """Default implementation propagates to input"""
        if self.input is not None:
            self.input.set_limit(limit)



"""
Pixeltable [UDFs](https://pixeltable.readme.io/docs/user-defined-functions-udfs) for `AudioType`.
"""

import pixeltable as pxt
from pixeltable.utils.code import local_public_names


@pxt.udf(is_method=True)
def get_metadata(audio: pxt.Audio) -> dict:
    """
    Gets various metadata associated with an audio file and returns it as a dictionary.

    Args:
        audio: The audio to get metadata for.

    Returns:
        A `dict` such as the following:

            ```json
            {
                'size': 2568827,
                'streams': [
                    {
                        'type': 'audio',
                        'frames': 0,
                        'duration': 2646000,
                        'metadata': {},
                        'time_base': 2.2675736961451248e-05,
                        'codec_context': {
                            'name': 'flac',
                            'profile': None,
                            'channels': 1,
                            'codec_tag': '\\x00\\x00\\x00\\x00',
                        },
                        'duration_seconds': 60.0,
                    }
                ],
                'bit_rate': 342510,
                'metadata': {'encoder': 'Lavf61.1.100'},
                'bit_exact': False,
            }
            ```

    Examples:
        Extract metadata for files in the `audio_col` column of the table `tbl`:

        >>> tbl.select(tbl.audio_col.get_metadata()).collect()
    """
    return pxt.functions.video._get_metadata(audio)


__all__ = local_public_names(__name__)


def __dir__() -> list[str]:
    return __all__



from pathlib import Path
from typing import TYPE_CHECKING, Any, Optional

import pixeltable as pxt
import pixeltable.exceptions as excs
from pixeltable.env import Env
from pixeltable.utils.code import local_public_names

if TYPE_CHECKING:
    import llama_cpp


@pxt.udf
def create_chat_completion(
    messages: list[dict],
    *,
    model_path: Optional[str] = None,
    repo_id: Optional[str] = None,
    repo_filename: Optional[str] = None,
    model_kwargs: Optional[dict[str, Any]] = None,
) -> dict:
    """
    Generate a chat completion from a list of messages.

    The model can be specified either as a local path, or as a repo_id and repo_filename that reference a pretrained
    model on the Hugging Face model hub. Exactly one of `model_path` or `repo_id` must be provided; if `model_path`
    is provided, then an optional `repo_filename` can also be specified.

    For additional details, see the
    [llama_cpp create_chat_completions documentation](https://llama-cpp-python.readthedocs.io/en/latest/api-reference/#llama_cpp.Llama.create_chat_completion).

    Args:
        messages: A list of messages to generate a response for.
        model_path: Path to the model (if using a local model).
        repo_id: The Hugging Face model repo id (if using a pretrained model).
        repo_filename: A filename or glob pattern to match the model file in the repo (optional, if using a
            pretrained model).
        model_kwargs: Additional keyword args for the llama_cpp `create_chat_completions` API, such as `max_tokens`,
            `temperature`, `top_p`, and `top_k`. For details, see the
            [llama_cpp create_chat_completions documentation](https://llama-cpp-python.readthedocs.io/en/latest/api-reference/#llama_cpp.Llama.create_chat_completion).
    """
    Env.get().require_package('llama_cpp', min_version=[0, 3, 1])

    if model_kwargs is None:
        model_kwargs = {}

    if (model_path is None) == (repo_id is None):
        raise excs.Error('Exactly one of `model_path` or `repo_id` must be provided.')
    if (repo_id is None) and (repo_filename is not None):
        raise excs.Error('`repo_filename` can only be provided along with `repo_id`.')

    n_gpu_layers = -1 if _is_gpu_available() else 0  # 0 = CPU only, -1 = offload all layers to GPU

    if model_path is not None:
        llm = _lookup_local_model(model_path, n_gpu_layers)
    else:
        Env.get().require_package('huggingface_hub')
        llm = _lookup_pretrained_model(repo_id, repo_filename, n_gpu_layers)
    return llm.create_chat_completion(messages, **model_kwargs)  # type: ignore


def _is_gpu_available() -> bool:
    import llama_cpp

    global _IS_GPU_AVAILABLE  # noqa: PLW0603
    if _IS_GPU_AVAILABLE is None:
        llama_cpp_path = Path(llama_cpp.__file__).parent
        lib = llama_cpp.llama_cpp.load_shared_library('llama', llama_cpp_path / 'lib')
        _IS_GPU_AVAILABLE = bool(lib.llama_supports_gpu_offload())

    return _IS_GPU_AVAILABLE


def _lookup_local_model(model_path: str, n_gpu_layers: int) -> 'llama_cpp.Llama':
    import llama_cpp

    key = (model_path, None, n_gpu_layers)
    if key not in _model_cache:
        llm = llama_cpp.Llama(model_path, n_gpu_layers=n_gpu_layers, verbose=False)
        _model_cache[key] = llm
    return _model_cache[key]


def _lookup_pretrained_model(repo_id: str, filename: Optional[str], n_gpu_layers: int) -> 'llama_cpp.Llama':
    import llama_cpp

    key = (repo_id, filename, n_gpu_layers)
    if key not in _model_cache:
        llm = llama_cpp.Llama.from_pretrained(
            repo_id=repo_id, filename=filename, n_gpu_layers=n_gpu_layers, verbose=False
        )
        _model_cache[key] = llm
    return _model_cache[key]


_model_cache: dict[tuple[str, str, int], Any] = {}
_IS_GPU_AVAILABLE: Optional[bool] = None


__all__ = local_public_names(__name__)


def __dir__() -> list[str]:
    return __all__



from __future__ import annotations

from typing import Any, Optional, Union

import sqlalchemy as sql

from .data_row import DataRow
from .expr import Expr
from .globals import print_slice
from .row_builder import RowBuilder
from .sql_element_cache import SqlElementCache


class ArraySlice(Expr):
    """
    Slice operation on an array, eg, t.array_col[:, 1:2].
    """

    def __init__(self, arr: Expr, index: tuple[Union[int, slice], ...]):
        assert arr.col_type.is_array_type()
        # determine result type
        super().__init__(arr.col_type)
        self.components = [arr]
        self.index = index
        self.id = self._create_id()

    def __repr__(self) -> str:
        index_strs: list[str] = []
        for el in self.index:
            if isinstance(el, int):
                index_strs.append(str(el))
            if isinstance(el, slice):
                index_strs.append(print_slice(el))
        return f'{self._array}[{", ".join(index_strs)}]'

    @property
    def _array(self) -> Expr:
        return self.components[0]

    def _equals(self, other: ArraySlice) -> bool:
        return self.index == other.index

    def _id_attrs(self) -> list[tuple[str, Any]]:
        return [*super()._id_attrs(), ('index', self.index)]

    def sql_expr(self, _: SqlElementCache) -> Optional[sql.ColumnElement]:
        return None

    def eval(self, data_row: DataRow, row_builder: RowBuilder) -> None:
        val = data_row[self._array.slot_idx]
        data_row[self.slot_idx] = val[self.index]

    def _as_dict(self) -> dict:
        index: list[Any] = []
        for el in self.index:
            if isinstance(el, slice):
                index.append([el.start, el.stop, el.step])
            else:
                index.append(el)
        return {'index': index, **super()._as_dict()}

    @classmethod
    def _from_dict(cls, d: dict, components: list[Expr]) -> ArraySlice:
        assert 'index' in d
        index = []
        for el in d['index']:
            if isinstance(el, list):
                index.append(slice(el[0], el[1], el[2]))
            else:
                index.append(el)
        return cls(components[0], tuple(index))



from __future__ import annotations

import logging
import warnings
from textwrap import dedent
from typing import TYPE_CHECKING, Any, Optional

import sqlalchemy as sql

import pixeltable.exceptions as excs
import pixeltable.type_system as ts
from pixeltable import exprs

from .globals import MediaValidation, is_valid_identifier

if TYPE_CHECKING:
    from .table_version import TableVersion
    from .table_version_handle import ColumnHandle
    from .table_version_path import TableVersionPath

_logger = logging.getLogger('pixeltable')


class Column:
    """Representation of a column in the schema of a Table/DataFrame.

    A Column contains all the metadata necessary for executing queries and updates against a particular version of a
    table/view.
    """

    name: str
    id: Optional[int]
    col_type: ts.ColumnType
    stored: bool
    is_pk: bool
    _media_validation: Optional[MediaValidation]  # if not set, TableVersion.media_validation applies
    schema_version_add: Optional[int]
    schema_version_drop: Optional[int]
    _stores_cellmd: Optional[bool]
    sa_col: Optional[sql.schema.Column]
    sa_col_type: Optional[sql.sqltypes.TypeEngine]
    sa_cellmd_col: Optional[sql.schema.Column]  # JSON metadata for the cell, e.g. errortype, errormsg for media columns
    _value_expr: Optional[exprs.Expr]
    value_expr_dict: Optional[dict[str, Any]]
    # we store a TableVersion here, not a TableVersionHandle, because this column is owned by that TableVersion instance
    # (re-resolving it later to a different instance doesn't make sense)
    tbl: Optional[TableVersion]

    def __init__(
        self,
        name: Optional[str],
        col_type: Optional[ts.ColumnType] = None,
        computed_with: Optional[exprs.Expr] = None,
        is_pk: bool = False,
        stored: bool = True,
        media_validation: Optional[MediaValidation] = None,
        col_id: Optional[int] = None,
        schema_version_add: Optional[int] = None,
        schema_version_drop: Optional[int] = None,
        sa_col_type: Optional[sql.sqltypes.TypeEngine] = None,
        stores_cellmd: Optional[bool] = None,
        value_expr_dict: Optional[dict[str, Any]] = None,
        tbl: Optional[TableVersion] = None,
    ):
        """Column constructor.

        Args:
            name: column name; None for system columns (eg, index columns)
            col_type: column type; can be None if the type can be derived from ``computed_with``
            computed_with: an Expr that computes the column value
            is_pk: if True, this column is part of the primary key
            stored: determines whether a computed column is present in the stored table or recomputed on demand
            col_id: column ID (only used internally)

        Computed columns: those have a non-None ``computed_with`` argument
        - when constructed by the user: ``computed_with`` was constructed explicitly and is passed in;
          col_type is None
        - when loaded from md store: ``computed_with`` is set and col_type is set

        ``stored`` (only valid for computed columns):
        - if True: the column is present in the stored table
        - if False: the column is not present in the stored table and recomputed during a query
        - if None: the system chooses for you (at present, this is always False, but this may change in the future)
        """
        if name is not None and not is_valid_identifier(name):
            raise excs.Error(f"Invalid column name: '{name}'")
        self.name = name
        self.tbl = tbl
        if col_type is None and computed_with is None:
            raise excs.Error(f'Column `{name}`: col_type is required if computed_with is not specified')

        self._value_expr: Optional[exprs.Expr] = None
        self.value_expr_dict = value_expr_dict
        if computed_with is not None:
            value_expr = exprs.Expr.from_object(computed_with)
            if value_expr is None:
                # TODO: this shouldn't be a user-facing error
                raise excs.Error(
                    f'Column {name}: computed_with needs to be a valid Pixeltable expression, '
                    f'but it is a {type(computed_with)}'
                )
            else:
                self._value_expr = value_expr.copy()
                self.col_type = self._value_expr.col_type
        if self._value_expr is not None and self.value_expr_dict is None:
            self.value_expr_dict = self._value_expr.as_dict()

        if col_type is not None:
            self.col_type = col_type
        assert self.col_type is not None

        self.stored = stored
        # self.dependent_cols = set()  # cols with value_exprs that reference us; set by TableVersion
        self.id = col_id
        self.is_pk = is_pk
        self._media_validation = media_validation
        self.schema_version_add = schema_version_add
        self.schema_version_drop = schema_version_drop

        self._stores_cellmd = stores_cellmd

        # column in the stored table for the values of this Column
        self.sa_col = None
        self.sa_col_type = sa_col_type

        # computed cols also have storage columns for the exception string and type
        self.sa_cellmd_col = None

    def init_value_expr(self) -> None:
        from pixeltable import exprs

        if self._value_expr is not None or self.value_expr_dict is None:
            return
        self._value_expr = exprs.Expr.from_dict(self.value_expr_dict)
        self._value_expr.bind_rel_paths()
        if not self._value_expr.is_valid:
            message = (
                dedent(
                    f"""
                    The computed column {self.name!r} in table {self.tbl.name!r} is no longer valid.
                    {{validation_error}}
                    You can continue to query existing data from this column, but evaluating it on new data will raise an error.
                    """  # noqa: E501
                )
                .strip()
                .format(validation_error=self._value_expr.validation_error)
            )
            warnings.warn(message, category=excs.PixeltableWarning, stacklevel=2)

    @property
    def handle(self) -> 'ColumnHandle':
        """Returns a ColumnHandle for this Column."""
        from .table_version_handle import ColumnHandle

        assert self.tbl is not None
        assert self.id is not None
        return ColumnHandle(self.tbl.handle, self.id)

    @property
    def value_expr(self) -> Optional[exprs.Expr]:
        assert self.value_expr_dict is None or self._value_expr is not None
        return self._value_expr

    def set_value_expr(self, value_expr: exprs.Expr) -> None:
        self._value_expr = value_expr
        self.value_expr_dict = self._value_expr.as_dict()

    def check_value_expr(self) -> None:
        assert self._value_expr is not None
        if self.stored == False and self.is_computed and self.has_window_fn_call():
            raise excs.Error(
                f'Column {self.name}: stored={self.stored} not supported for columns computed with window functions:'
                f'\n{self.value_expr}'
            )

    def has_window_fn_call(self) -> bool:
        if self.value_expr is None:
            return False
        from pixeltable import exprs

        window_fn_calls = list(
            self.value_expr.subexprs(filter=lambda e: isinstance(e, exprs.FunctionCall) and e.is_window_fn_call)
        )
        return len(window_fn_calls) > 0

    # TODO: This should be moved out of `Column` (its presence in `Column` doesn't anticipate indices being defined on
    #     multiple dependents)
    def get_idx_info(self, reference_tbl: Optional['TableVersionPath'] = None) -> dict[str, 'TableVersion.IndexInfo']:
        assert self.tbl is not None
        tbl = reference_tbl.tbl_version.get() if reference_tbl is not None else self.tbl
        return {name: info for name, info in tbl.idxs_by_name.items() if info.col == self}

    @property
    def is_computed(self) -> bool:
        return self._value_expr is not None or self.value_expr_dict is not None

    @property
    def is_stored(self) -> bool:
        """Returns True if column is materialized in the stored table."""
        assert self.stored is not None
        return self.stored

    @property
    def stores_cellmd(self) -> bool:
        """True if this column also stores error information."""
        # default: record errors for computed and media columns
        if self._stores_cellmd is not None:
            return self._stores_cellmd
        return self.is_stored and (self.is_computed or self.col_type.is_media_type())

    @property
    def qualified_name(self) -> str:
        assert self.tbl is not None
        return f'{self.tbl.name}.{self.name}'

    @property
    def media_validation(self) -> MediaValidation:
        if self._media_validation is not None:
            return self._media_validation
        assert self.tbl is not None
        return self.tbl.media_validation

    @property
    def is_required_for_insert(self) -> bool:
        """Returns True if column is required when inserting rows."""
        return not self.col_type.nullable and not self.is_computed

    def source(self) -> None:
        """
        If this is a computed col and the top-level expr is a function call, print the source, if possible.
        """
        from pixeltable import exprs

        if self.value_expr is None or not isinstance(self.value_expr, exprs.FunctionCall):
            return
        self.value_expr.fn.source()

    def create_sa_cols(self) -> None:
        """
        These need to be recreated for every new table schema version.
        """
        assert self.is_stored
        # all storage columns are nullable (we deal with null errors in Pixeltable directly)
        self.sa_col = sql.Column(self.store_name(), self.get_sa_col_type(), nullable=True)
        if self.stores_cellmd:
            # JSON metadata for the cell, e.g. errortype, errormsg for media columns
            self.sa_cellmd_col = sql.Column(self.cellmd_store_name(), self.sa_cellmd_type(), nullable=True)

    def get_sa_col_type(self) -> sql.sqltypes.TypeEngine:
        return self.col_type.to_sa_type() if self.sa_col_type is None else self.sa_col_type

    @classmethod
    def cellmd_type(cls) -> ts.ColumnType:
        return ts.JsonType(nullable=True)

    @classmethod
    def sa_cellmd_type(cls) -> sql.sqltypes.TypeEngine:
        return cls.cellmd_type().to_sa_type()

    def store_name(self) -> str:
        assert self.id is not None
        assert self.is_stored
        return f'col_{self.id}'

    def cellmd_store_name(self) -> str:
        return f'{self.store_name()}_cellmd'

    def __str__(self) -> str:
        return f'{self.name}: {self.col_type}'

    def __repr__(self) -> str:
        return f'Column({self.id!r}, {self.name!r}, tbl={self.tbl.name!r})'

    def __hash__(self) -> int:
        # TODO(aaron-siegel): This and __eq__ do not capture the table version. We need to rethink the Column
        #     abstraction (perhaps separating out the version-dependent properties into a different abstraction).
        assert self.tbl is not None
        return hash((self.tbl.id, self.id))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Column):
            return False
        assert self.tbl is not None
        assert other.tbl is not None
        return self.tbl.id == other.tbl.id and self.id == other.id



import dataclasses
import enum
import logging
from typing import Any, ClassVar, Iterable, Iterator, Optional, Union

import ftfy

from pixeltable.env import Env
from pixeltable.exceptions import Error
from pixeltable.type_system import ColumnType, DocumentType, IntType, JsonType, StringType
from pixeltable.utils.documents import get_document_handle

from .base import ComponentIterator

_logger = logging.getLogger('pixeltable')


class ChunkMetadata(enum.Enum):
    TITLE = 1
    HEADING = 2
    SOURCELINE = 3
    PAGE = 4
    BOUNDING_BOX = 5


class Separator(enum.Enum):
    HEADING = 1
    PARAGRAPH = 2
    SENTENCE = 3
    TOKEN_LIMIT = 4
    CHAR_LIMIT = 5
    PAGE = 6


@dataclasses.dataclass
class DocumentSectionMetadata:
    """Metadata for a subsection of a document (ie, a structural element like a heading or paragraph)"""

    # html and markdown metadata
    sourceline: Optional[int] = None
    # the stack of headings up to the most recently observed one;
    # eg, if the most recent one was an h2, 'headings' would contain keys 1 and 2, but nothing below that
    heading: Optional[dict[str, str]] = None

    # pdf-specific metadata
    page: Optional[int] = None
    # bounding box as an {x1, y1, x2, y2} dictionary
    bounding_box: Optional[dict[str, float]] = None


@dataclasses.dataclass
class DocumentSection:
    """A single document chunk, according to some of the splitting criteria"""

    text: Optional[str]
    metadata: Optional[DocumentSectionMetadata]


def _parse_separators(separators: str) -> list[Separator]:
    ret = []
    for s in separators.split(','):
        clean_s = s.strip().upper()
        if not clean_s:
            continue
        if clean_s not in Separator.__members__:
            raise Error(
                f'Invalid separator: `{s.strip()}`. Valid separators are: {", ".join(Separator.__members__).lower()}'
            )
        ret.append(Separator[clean_s])
    return ret


def _parse_metadata(metadata: str) -> list[ChunkMetadata]:
    ret = []
    for m in metadata.split(','):
        clean_m = m.strip().upper()
        if not clean_m:
            continue
        if clean_m not in ChunkMetadata.__members__:
            raise Error(
                f'Invalid metadata: `{m.strip()}`. Valid metadata are: {", ".join(ChunkMetadata.__members__).lower()}'
            )
        ret.append(ChunkMetadata[clean_m])
    return ret


_HTML_HEADINGS = {'h1', 'h2', 'h3', 'h4', 'h5', 'h6'}


class DocumentSplitter(ComponentIterator):
    """Iterator over chunks of a document. The document is chunked according to the specified `separators`.

    The iterator yields a `text` field containing the text of the chunk, and it may also
    include additional metadata fields if specified in the `metadata` parameter, as explained below.

    Chunked text will be cleaned with `ftfy.fix_text` to fix up common problems with unicode sequences.
    """

    METADATA_COLUMN_TYPES: ClassVar[dict[ChunkMetadata, ColumnType]] = {
        ChunkMetadata.TITLE: StringType(nullable=True),
        ChunkMetadata.HEADING: JsonType(nullable=True),
        ChunkMetadata.SOURCELINE: IntType(nullable=True),
        ChunkMetadata.PAGE: IntType(nullable=True),
        ChunkMetadata.BOUNDING_BOX: JsonType(nullable=True),
    }

    def __init__(
        self,
        document: str,
        *,
        separators: str,
        limit: Optional[int] = None,
        overlap: Optional[int] = None,
        metadata: str = '',
        html_skip_tags: Optional[list[str]] = None,
        tiktoken_encoding: Optional[str] = 'cl100k_base',
        tiktoken_target_model: Optional[str] = None,
    ):
        """Init method for `DocumentSplitter` class.

        Args:
            separators: separators to use to chunk the document. Options are:
                 `'heading'`, `'paragraph'`, `'sentence'`, `'token_limit'`, `'char_limit'`, `'page'`.
                 This may be a comma-separated string, e.g., `'heading,token_limit'`.
            limit: the maximum number of tokens or characters in each chunk, if `'token_limit'`
                 or `'char_limit'` is specified.
            metadata: additional metadata fields to include in the output. Options are:
                 `'title'`, `'heading'` (HTML and Markdown), `'sourceline'` (HTML), `'page'` (PDF), `'bounding_box'`
                 (PDF). The input may be a comma-separated string, e.g., `'title,heading,sourceline'`.
        """
        if html_skip_tags is None:
            html_skip_tags = ['nav']
        self._doc_handle = get_document_handle(document)
        assert self._doc_handle is not None
        # calling the output_schema method to validate the input arguments
        self.output_schema(separators=separators, metadata=metadata, limit=limit, overlap=overlap)
        self._separators = _parse_separators(separators)
        self._metadata_fields = _parse_metadata(metadata)
        if self._doc_handle.bs_doc is not None:
            title = self._doc_handle.bs_doc.title
            if title is None:
                self._doc_title = ''
            else:
                self._doc_title = ftfy.fix_text(title.get_text().strip())
        else:
            self._doc_title = ''
        self._limit = 0 if limit is None else limit
        self._skip_tags = html_skip_tags
        self._overlap = 0 if overlap is None else overlap
        self._tiktoken_encoding = tiktoken_encoding
        self._tiktoken_target_model = tiktoken_target_model

        # set up processing pipeline
        if self._doc_handle.format == DocumentType.DocumentFormat.HTML:
            assert self._doc_handle.bs_doc is not None
            self._sections = self._html_sections()
        elif self._doc_handle.format == DocumentType.DocumentFormat.MD:
            assert self._doc_handle.md_ast is not None
            self._sections = self._markdown_sections()
        elif self._doc_handle.format == DocumentType.DocumentFormat.PDF:
            assert self._doc_handle.pdf_doc is not None
            self._sections = self._pdf_sections()
        elif self._doc_handle.format == DocumentType.DocumentFormat.TXT:
            assert self._doc_handle.txt_doc is not None
            self._sections = self._txt_sections()
        else:
            raise AssertionError(f'Unsupported document format: {self._doc_handle.format}')

        if Separator.SENTENCE in self._separators:
            self._sections = self._sentence_sections(self._sections)
        if Separator.TOKEN_LIMIT in self._separators:
            self._sections = self._token_chunks(self._sections)
        if Separator.CHAR_LIMIT in self._separators:
            self._sections = self._char_chunks(self._sections)

    @classmethod
    def input_schema(cls) -> dict[str, ColumnType]:
        return {
            'document': DocumentType(nullable=False),
            'separators': StringType(nullable=False),
            'metadata': StringType(nullable=False),
            'limit': IntType(nullable=True),
            'overlap': IntType(nullable=True),
            'skip_tags': StringType(nullable=True),
            'tiktoken_encoding': StringType(nullable=True),
            'tiktoken_target_model': StringType(nullable=True),
        }

    @classmethod
    def output_schema(cls, *args: Any, **kwargs: Any) -> tuple[dict[str, ColumnType], list[str]]:
        schema: dict[str, ColumnType] = {'text': StringType()}
        md_fields = _parse_metadata(kwargs['metadata']) if 'metadata' in kwargs else []

        for md_field in md_fields:
            schema[md_field.name.lower()] = cls.METADATA_COLUMN_TYPES[md_field]

        assert 'separators' in kwargs
        separators = _parse_separators(kwargs['separators'])

        limit = kwargs.get('limit')
        overlap = kwargs.get('overlap')

        if limit is not None or overlap is not None:
            if Separator.TOKEN_LIMIT not in separators and Separator.CHAR_LIMIT not in separators:
                raise Error('limit/overlap requires the "token_limit" or "char_limit" separator')
            if limit is not None and limit <= 0:
                raise Error('"limit" must be an integer > 0')
            if overlap is not None and overlap < 0:
                raise Error('"overlap" must be an integer >= 0')
        if Separator.TOKEN_LIMIT in separators or Separator.CHAR_LIMIT in separators:
            if Separator.TOKEN_LIMIT in separators and Separator.CHAR_LIMIT in separators:
                raise Error('Cannot specify both "token_limit" and "char_limit" separators')
            if kwargs.get('limit') is None:
                raise Error('limit is required with "token_limit"/"char_limit" separators')

        # check dependencies at the end
        if Separator.SENTENCE in separators:
            _ = Env.get().spacy_nlp
        if Separator.TOKEN_LIMIT in separators:
            Env.get().require_package('tiktoken')

        return schema, []

    def __next__(self) -> dict[str, Any]:
        while True:
            section = next(self._sections)
            if section.text is None:
                continue
            result: dict[str, Any] = {'text': section.text}
            for md_field in self._metadata_fields:
                if md_field == ChunkMetadata.TITLE:
                    result[md_field.name.lower()] = self._doc_title
                elif md_field == ChunkMetadata.HEADING:
                    result[md_field.name.lower()] = section.metadata.heading
                elif md_field == ChunkMetadata.SOURCELINE:
                    result[md_field.name.lower()] = section.metadata.sourceline
                elif md_field == ChunkMetadata.PAGE:
                    result[md_field.name.lower()] = section.metadata.page
                elif md_field == ChunkMetadata.BOUNDING_BOX:
                    result[md_field.name.lower()] = section.metadata.bounding_box
            return result

    def _html_sections(self) -> Iterator[DocumentSection]:
        """Create DocumentSections reflecting the html-specific separators"""
        import bs4

        emit_on_paragraph = Separator.PARAGRAPH in self._separators or Separator.SENTENCE in self._separators
        emit_on_heading = Separator.HEADING in self._separators or emit_on_paragraph
        # current state
        accumulated_text: list[str] = []  # currently accumulated text
        # accumulate pieces then join before emit to avoid quadratic complexity of string concatenation

        headings: dict[str, str] = {}  # current state of observed headings (level -> text)
        sourceline = 0  # most recently seen sourceline

        def update_metadata(el: bs4.Tag) -> None:
            # update current state
            nonlocal headings, sourceline
            sourceline = el.sourceline
            if el.name in _HTML_HEADINGS:
                # remove the previously seen lower levels
                lower_levels = [lv for lv in headings if lv > el.name]
                for lv in lower_levels:
                    del headings[lv]
                headings[el.name] = el.get_text().strip()

        def emit() -> Iterator[DocumentSection]:
            nonlocal accumulated_text, headings, sourceline
            if len(accumulated_text) > 0:
                md = DocumentSectionMetadata(sourceline=sourceline, heading=headings.copy())
                full_text = ' '.join(accumulated_text)
                full_text = ftfy.fix_text(full_text)
                yield DocumentSection(text=full_text, metadata=md)
                accumulated_text = []

        def process_element(el: Union[bs4.element.Tag, bs4.NavigableString]) -> Iterator[DocumentSection]:
            # process the element and emit sections as necessary
            nonlocal accumulated_text, headings, sourceline, emit_on_heading, emit_on_paragraph

            if el.name in self._skip_tags:
                return

            if isinstance(el, bs4.NavigableString):
                # accumulate text until we see a tag we care about
                text = el.get_text().strip()
                if len(text) > 0:
                    accumulated_text.append(text)
                return

            if el.name in _HTML_HEADINGS:
                if emit_on_heading:
                    yield from emit()
                update_metadata(el)
            elif el.name == 'p':
                if emit_on_paragraph:
                    yield from emit()
                update_metadata(el)
            for child in el.children:
                assert isinstance(child, (bs4.element.Tag, bs4.NavigableString)), type(el)
                yield from process_element(child)

        yield from process_element(self._doc_handle.bs_doc)
        yield from emit()

    def _markdown_sections(self) -> Iterator[DocumentSection]:
        """Create DocumentSections reflecting the html-specific separators"""
        assert self._doc_handle.md_ast is not None
        emit_on_paragraph = Separator.PARAGRAPH in self._separators or Separator.SENTENCE in self._separators
        emit_on_heading = Separator.HEADING in self._separators or emit_on_paragraph
        # current state
        accumulated_text: list[str] = []  # currently accumulated text
        # accumulate pieces then join before emit to avoid quadratic complexity of string concatenation
        headings: dict[str, str] = {}  # current state of observed headings (level -> text)

        def update_headings(heading: dict) -> None:
            # update current state
            nonlocal headings
            assert 'type' in heading and heading['type'] == 'heading'
            lint = heading['attrs']['level']
            level = f'h{lint}'
            text = heading['children'][0]['raw'].strip()
            # remove the previously seen lower levels
            lower_levels = [lv for lv in headings if lv > level]
            for lv in lower_levels:
                del headings[lv]
            headings[level] = text

        def emit() -> Iterator[DocumentSection]:
            nonlocal accumulated_text, headings
            if len(accumulated_text) > 0:
                metadata = DocumentSectionMetadata(sourceline=0, heading=headings.copy())
                yield DocumentSection(text=ftfy.fix_text(' '.join(accumulated_text)), metadata=metadata)
                accumulated_text = []

        def process_element(el: dict) -> Iterator[DocumentSection]:
            # process the element and emit sections as necessary
            nonlocal accumulated_text, headings, emit_on_heading, emit_on_paragraph
            assert 'type' in el

            if el['type'] == 'text':
                # accumulate text until we see a separator element
                text = el['raw'].strip()
                if len(text) > 0:
                    accumulated_text.append(text)
                return

            if el['type'] == 'heading':
                if emit_on_heading:
                    yield from emit()
                update_headings(el)
            elif el['type'] == 'paragraph':
                if emit_on_paragraph:
                    yield from emit()
            if 'children' not in el:
                return
            for child in el['children']:
                yield from process_element(child)

        for el in self._doc_handle.md_ast:
            yield from process_element(el)
        yield from emit()

    def _pdf_sections(self) -> Iterator[DocumentSection]:
        """Create DocumentSections reflecting the pdf-specific separators"""
        import fitz  # type: ignore[import-untyped]

        doc: fitz.Document = self._doc_handle.pdf_doc
        assert doc is not None

        emit_on_paragraph = Separator.PARAGRAPH in self._separators or Separator.SENTENCE in self._separators
        emit_on_page = Separator.PAGE in self._separators or emit_on_paragraph

        accumulated_text = []  # invariant: all elements are ftfy clean and non-empty

        def _add_cleaned_text(raw_text: str) -> None:
            fixed = ftfy.fix_text(raw_text)
            if fixed:
                accumulated_text.append(fixed)

        def _emit_text() -> str:
            full_text = ''.join(accumulated_text)
            accumulated_text.clear()
            return full_text

        for page_number, page in enumerate(doc.pages()):
            for block in page.get_text('blocks'):
                # there is no concept of paragraph in pdf, block is the closest thing
                # we can get (eg a paragraph in text may cut across pages)
                # see pymupdf docs https://pymupdf.readthedocs.io/en/latest/app1.html
                # other libraries like pdfminer also lack an explicit paragraph concept
                x1, y1, x2, y2, text, _, _ = block
                _add_cleaned_text(text)
                if accumulated_text and emit_on_paragraph:
                    bbox = {'x1': x1, 'y1': y1, 'x2': x2, 'y2': y2}
                    metadata = DocumentSectionMetadata(page=page_number, bounding_box=bbox)
                    yield DocumentSection(text=_emit_text(), metadata=metadata)

            if accumulated_text and emit_on_page and not emit_on_paragraph:
                yield DocumentSection(text=_emit_text(), metadata=DocumentSectionMetadata(page=page_number))
                accumulated_text = []

        if accumulated_text and not emit_on_page:
            yield DocumentSection(text=_emit_text(), metadata=DocumentSectionMetadata())

    def _txt_sections(self) -> Iterator[DocumentSection]:
        """Create DocumentSections for text files.

        Currently, it returns the entire text as a single section.
        TODO: Add support for paragraphs.
        """
        assert self._doc_handle.txt_doc is not None
        yield DocumentSection(text=ftfy.fix_text(self._doc_handle.txt_doc), metadata=DocumentSectionMetadata())

    def _sentence_sections(self, input_sections: Iterable[DocumentSection]) -> Iterator[DocumentSection]:
        """Split the input sections into sentences"""
        for section in input_sections:
            if section.text is not None:
                doc = Env.get().spacy_nlp(section.text)
                for sent in doc.sents:
                    yield DocumentSection(text=sent.text, metadata=section.metadata)

    def _token_chunks(self, input: Iterable[DocumentSection]) -> Iterator[DocumentSection]:
        import tiktoken

        if self._tiktoken_target_model is not None:
            encoding = tiktoken.encoding_for_model(self._tiktoken_target_model)
        else:
            encoding = tiktoken.get_encoding(self._tiktoken_encoding)
        assert self._limit > 0 and self._overlap >= 0

        for section in input:
            if section.text is None:
                continue
            tokens = encoding.encode(section.text)
            start_idx = 0
            text = None
            while start_idx < len(tokens):
                end_idx = min(start_idx + self._limit, len(tokens))
                while end_idx > start_idx:
                    # find a cutoff point that doesn't cut in the middle of utf8 multi-byte sequences
                    try:
                        # check that the truncated data can be properly decoded
                        text = encoding.decode(tokens[start_idx:end_idx], errors='strict')
                        break
                    except UnicodeDecodeError:
                        # we split the token array at a point where the utf8 encoding is broken
                        end_idx -= 1

                assert end_idx > start_idx
                assert text
                yield DocumentSection(text=text, metadata=section.metadata)
                start_idx = max(start_idx + 1, end_idx - self._overlap)  # ensure we make progress

    def _char_chunks(self, input: Iterable[DocumentSection]) -> Iterator[DocumentSection]:
        for section in input:
            if section.text is None:
                continue
            start_idx = 0
            while start_idx < len(section.text):
                end_idx = min(start_idx + self._limit, len(section.text))
                text = section.text[start_idx:end_idx]
                yield DocumentSection(text=text, metadata=section.metadata)
                start_idx += self._limit - self._overlap

    def close(self) -> None:
        pass

    def set_pos(self, pos: int) -> None:
        pass



from __future__ import annotations

from typing import Any, Iterable, Optional

import numpy as np
import sqlalchemy as sql

import pixeltable.exceptions as excs
import pixeltable.type_system as ts

from .data_row import DataRow
from .expr import Expr
from .literal import Literal
from .row_builder import RowBuilder
from .sql_element_cache import SqlElementCache


class InlineArray(Expr):
    """
    Array 'literal' which can use Exprs as values.
    """

    def __init__(self, elements: Iterable):
        exprs = []
        for el in elements:
            if isinstance(el, Expr):
                exprs.append(el)
            elif isinstance(el, (list, tuple)):
                exprs.append(Expr.from_array(el))
            else:
                exprs.append(Literal(el))

        inferred_element_type: Optional[ts.ColumnType] = ts.InvalidType()
        for i, expr in enumerate(exprs):
            supertype = inferred_element_type.supertype(expr.col_type)
            if supertype is None:
                raise excs.Error(
                    f'Could not infer element type of array: element of type `{expr.col_type}` at index {i} '
                    f'is not compatible with type `{inferred_element_type}` of preceding elements'
                )
            inferred_element_type = supertype

        if inferred_element_type.is_scalar_type():
            col_type = ts.ArrayType((len(exprs),), inferred_element_type)
        elif inferred_element_type.is_array_type():
            assert isinstance(inferred_element_type, ts.ArrayType)
            col_type = ts.ArrayType(
                (len(exprs), *inferred_element_type.shape), ts.ColumnType.make_type(inferred_element_type.dtype)
            )
        else:
            raise excs.Error(f'Element type is not a valid dtype for an array: {inferred_element_type}')

        super().__init__(col_type)
        self.components.extend(exprs)
        self.id = self._create_id()

    def __repr__(self) -> str:
        elem_strs = [str(expr) for expr in self.components]
        return f'[{", ".join(elem_strs)}]'

    def _equals(self, _: InlineArray) -> bool:
        return True  # Always true if components match

    def sql_expr(self, _: SqlElementCache) -> Optional[sql.ColumnElement]:
        return None

    def eval(self, data_row: DataRow, row_builder: RowBuilder) -> None:
        data_row[self.slot_idx] = np.array([data_row[el.slot_idx] for el in self.components])

    def _as_dict(self) -> dict:
        return super()._as_dict()

    @classmethod
    def _from_dict(cls, _: dict, components: list[Expr]) -> InlineArray:
        try:
            return cls(components)
        except excs.Error:
            # For legacy compatibility reasons, we need to try constructing as an `InlineList`.
            # This is because in schema versions <= 19, `InlineArray` was serialized incorrectly, and
            # there is no way to determine the correct expression type until the subexpressions are
            # loaded and their types are known.
            return InlineList(components)  # type: ignore[return-value]

    def as_literal(self) -> Optional[Literal]:
        assert isinstance(self.col_type, ts.ArrayType)
        if not all(isinstance(comp, Literal) for comp in self.components):
            return None
        return Literal(
            np.array([c.as_literal().val for c in self.components], dtype=self.col_type.numpy_dtype()), self.col_type
        )


class InlineList(Expr):
    """
    List 'literal' which can use Exprs as values.
    """

    def __init__(self, elements: Iterable):
        exprs = [Expr.from_object(el) for el in elements]

        json_schema = {
            'type': 'array',
            'prefixItems': [expr.col_type.to_json_schema() for expr in exprs],
            'items': False,  # No additional items (fixed length)
        }

        super().__init__(ts.JsonType(json_schema))
        self.components.extend(exprs)
        self.id = self._create_id()

    def __repr__(self) -> str:
        elem_strs = [str(expr) for expr in self.components]
        return f'[{", ".join(elem_strs)}]'

    def _equals(self, _: InlineList) -> bool:
        return True  # Always true if components match

    def sql_expr(self, _: SqlElementCache) -> Optional[sql.ColumnElement]:
        return None

    def eval(self, data_row: DataRow, _: RowBuilder) -> None:
        data_row[self.slot_idx] = [data_row[el.slot_idx] for el in self.components]

    def _as_dict(self) -> dict:
        return super()._as_dict()

    @classmethod
    def _from_dict(cls, _: dict, components: list[Expr]) -> InlineList:
        return cls(components)

    def as_literal(self) -> Optional[Literal]:
        if not all(isinstance(comp, Literal) for comp in self.components):
            return None
        return Literal([c.as_literal().val for c in self.components], self.col_type)


class InlineDict(Expr):
    """
    Dictionary 'literal' which can use Exprs as values.
    """

    keys: list[str]

    def __init__(self, d: dict[str, Any]):
        self.keys = []
        exprs: list[Expr] = []
        for key, val in d.items():
            if not isinstance(key, str):
                raise excs.Error(f'Dictionary requires string keys; {key} has type {type(key)}')
            self.keys.append(key)
            exprs.append(Expr.from_object(val))

        json_schema: Optional[dict[str, Any]]
        try:
            json_schema = {
                'type': 'object',
                'properties': {key: expr.col_type.to_json_schema() for key, expr in zip(self.keys, exprs)},
            }
        except excs.Error:
            # InlineDicts are used to store iterator arguments, which are not required to be valid JSON types,
            # so we can't always construct a valid schema.
            json_schema = None

        super().__init__(ts.JsonType(json_schema))
        self.components.extend(exprs)
        self.id = self._create_id()

    def __repr__(self) -> str:
        item_strs = [f"'{key}': {expr}" for key, expr in zip(self.keys, self.components)]
        return '{' + ', '.join(item_strs) + '}'

    def _equals(self, other: InlineDict) -> bool:
        # The dict values are just the components, which have already been checked
        return self.keys == other.keys

    def _id_attrs(self) -> list[tuple[str, Any]]:
        return [*super()._id_attrs(), ('keys', self.keys)]

    def sql_expr(self, _: SqlElementCache) -> Optional[sql.ColumnElement]:
        return None

    def eval(self, data_row: DataRow, _: RowBuilder) -> None:
        assert len(self.keys) == len(self.components)
        data_row[self.slot_idx] = {key: data_row[expr.slot_idx] for key, expr in zip(self.keys, self.components)}

    def to_kwargs(self) -> dict[str, Any]:
        """Deconstructs this expression into a dictionary by recursively unwrapping all Literals,
        InlineDicts, and InlineLists."""
        return InlineDict._to_kwarg_element(self)

    @classmethod
    def _to_kwarg_element(cls, expr: Expr) -> Any:
        if isinstance(expr, Literal):
            return expr.val
        if isinstance(expr, InlineDict):
            return {key: cls._to_kwarg_element(val) for key, val in zip(expr.keys, expr.components)}
        if isinstance(expr, InlineList):
            return [cls._to_kwarg_element(el) for el in expr.components]
        return expr

    def _as_dict(self) -> dict[str, Any]:
        return {'keys': self.keys, **super()._as_dict()}

    @classmethod
    def _from_dict(cls, d: dict, components: list[Expr]) -> InlineDict:
        assert 'keys' in d
        assert len(d['keys']) == len(components)
        arg = dict(zip(d['keys'], components))
        return InlineDict(arg)

    def as_literal(self) -> Optional[Literal]:
        if not all(isinstance(comp, Literal) for comp in self.components):
            return None
        return Literal(dict(zip(self.keys, (c.as_literal().val for c in self.components))), self.col_type)



from __future__ import annotations

import enum
import itertools
import logging
from dataclasses import dataclass
from typing import Optional
from uuid import UUID

import pixeltable.exceptions as excs

_logger = logging.getLogger('pixeltable')

# name of the position column in a component view
_POS_COLUMN_NAME = 'pos'
_ROWID_COLUMN_NAME = '_rowid'

# Set of symbols that are predefined in the `InsertableTable` class (and are therefore not allowed as column names).
# This will be populated lazily to avoid circular imports.
_PREDEF_SYMBOLS: Optional[set[str]] = None


@dataclass(frozen=True)
class QColumnId:
    """Qualified column id"""

    tbl_id: UUID
    col_id: int


class MediaValidation(enum.Enum):
    ON_READ = 0
    ON_WRITE = 1

    @classmethod
    def validated(cls, name: str, error_prefix: str) -> MediaValidation:
        try:
            return cls[name.upper()]
        except KeyError:
            val_strs = ', '.join(f'{s.lower()!r}' for s in cls.__members__)
            raise excs.Error(f'{error_prefix} must be one of: [{val_strs}]') from None


class IfExistsParam(enum.Enum):
    ERROR = 0
    IGNORE = 1
    REPLACE = 2
    REPLACE_FORCE = 3

    @classmethod
    def validated(cls, param_val: str, param_name: str) -> IfExistsParam:
        try:
            return cls[param_val.upper()]
        except KeyError:
            val_strs = ', '.join(f'{s.lower()!r}' for s in cls.__members__)
            raise excs.Error(f'{param_name} must be one of: [{val_strs}]') from None


class IfNotExistsParam(enum.Enum):
    ERROR = 0
    IGNORE = 1

    @classmethod
    def validated(cls, param_val: str, param_name: str) -> IfNotExistsParam:
        try:
            return cls[param_val.upper()]
        except KeyError:
            val_strs = ', '.join(f'{s.lower()!r}' for s in cls.__members__)
            raise excs.Error(f'{param_name} must be one of: [{val_strs}]') from None


def is_valid_identifier(name: str, allow_system_identifiers: bool = False) -> bool:
    return name.isidentifier() and (allow_system_identifiers or not name.startswith('_'))


def is_valid_path(path: str, empty_is_valid: bool, allow_system_paths: bool = False) -> bool:
    if path == '':
        return empty_is_valid
    return all(is_valid_identifier(part, allow_system_paths) for part in path.split('.'))


def is_system_column_name(name: str) -> bool:
    from pixeltable.catalog import InsertableTable, View

    global _PREDEF_SYMBOLS  # noqa: PLW0603
    if _PREDEF_SYMBOLS is None:
        _PREDEF_SYMBOLS = set(itertools.chain(dir(InsertableTable), dir(View)))
    return name == _POS_COLUMN_NAME or name in _PREDEF_SYMBOLS
